from PyQt5.QtCore import QPoint
from PyQt5.QtWidgets import QMenu, QWidget


def show( control: QWidget, *args ) -> object:
    menu = QMenu()
    
    if control.window().styleSheet():
        menu.setStyleSheet( control.window().styleSheet() )
    
    r = []
    
    for arg in args:
        r.append( menu.addAction( str( arg ) ) )
    
    selected = menu.exec_( control.mapToGlobal( QPoint( 0, control.height() ) ) )
    
    if selected is None:
        return None
    
    return args[r.index( selected )]



from mhelper_qt import qt_colour_helper, qt_gui_helper
from mhelper_qt.qt_colour_helper import Brushes, Colours, Pens
from mhelper_qt.qt_gui_helper import AnsiHtmlScheme, QtMutex, exceptToGui, exqtSlot



from typing import Union, List

import sys

from mhelper import string_helper, ansi
from mhelper.reflection_helper import TTristate


class DefaultLogger:
    def __init__( self ):
        self.print = self.__print_stderr
    
    
    def __print_stderr( self, text ):
        print( text, file = sys.stderr )


DEFAULT_LOGGER = DefaultLogger()


class Logger:
    """
    Logging simplified.
    """
    
    
    def __init__( self, name: Union[str, bool], enabled: bool = False ):
        """
        CONSTRUCTOR
        :param name:        Name of the logger 
        :param enabled:     A value for :property:`target` 
        """
        LOGGERS.append( self )
        self.name = name
        
        if name:
            self.__prefix = name + ": "
        else:
            self.__prefix = ""
        
        self.__target = None
        self.__enabled: TTristate = None
        self.target = enabled
        self.__skip = set()
    
    
    def pause( self, message = "PAUSED", *args, key = None ):
        if not self.__enabled:
            return
        
        if key and key in self.__skip:
            return
        
        message = self.format( message, *args )
        
        c = ansi.FORE_BRIGHT_WHITE + ansi.BACK_BLUE
        r = ansi.RESET
        
        if key is None:
            print( message + " " + c + "[C|D]" + r, end = "" )
        else:
            print( message + " " + c + "[C|I|D]" + r, end = "" )
        
        i = input()
        
        if i == "i" and key:
            self.__skip.add( key )
        elif i == "d":
            self.target = None
    
    
    @property
    def target( self ):
        """
        Gets or sets the target printer, which should accept one unnamed `str` parameter.
        The following are also accepted:
            `None` or `False`:  Disables printing
            `True`:             Prints to std-err.
        :return: 
        """
        return self.__target
    
    
    @property
    def enabled( self ) -> TTristate:
        return self.__enabled
    
    
    @target.setter
    def target( self, enabled: bool = None ):
        if enabled is True:
            self.__enabled = True
            self.__target = DEFAULT_LOGGER.print
        elif enabled is False or enabled is None:
            self.__enabled = False
            self.__target = None
        else:
            self.__enabled = None
            self.__target = enabled
    
    
    def __call__( self, *args ):
        if self.__target is None:
            return self
        
        self.print( self.format( *args ) )
        
        return self
    
    
    def format( self, *args ):
        if len( args ) == 1:
            return args[0]
        elif len( args ) > 1:
            vals = list( args[1:] )
            for i in range( len( vals ) ):
                v = vals[i]
                
                if type( v ) in (set, list, tuple, frozenset):
                    vals[i] = string_helper.format_array( v )
                
                vals[i] = "«" + str( vals[i] ) + "»"
            
            return args[0].format( *vals )
    
    
    def print( self, message ):
        if "\n" in message:
            for i, split in enumerate( message.split( "\n" ) ):
                self.print( ("\b\b: " if i != 0 else "") + split )
            
            return
        
        self.__target( ansi.DIM + self.__prefix + ansi.RESET + string_helper.highlight_quotes( message, "«", "»", ansi.FORE_YELLOW, ansi.RESET ) )
    
    
    def indent( self ):
        self.__prefix += "    "
        
        if len( self.__prefix ) > 100:
            from mhelper.exception_helper import ImplementationError
            raise ImplementationError( "Log indent too far." )
    
    
    def unindent( self ):
        self.__prefix = self.__prefix[:len( self.__prefix ) - 4]
    
    
    def __enter__( self ):
        self.indent()
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        self.unindent()


LOGGERS: List[Logger] = []
"""List of all loggers"""



"""
Helper functions for arrays, lists, iterable, etc.
"""

import warnings
from typing import List, Optional, Iterator, Tuple, Dict, Iterable, Union, TypeVar, Callable, cast

from mhelper.special_types import NOT_PROVIDED


T = TypeVar( "T" )
U = TypeVar( "U" )


def list_type( the_list ) -> type:
    """
    Determines the type of elements in a list

    Errors if the list doesn't contain any elements, or if the elements are of varying type

    :param the_list: List to check
    :return: Type of elements in the list
    """
    
    t = None
    
    for e in the_list:
        et = type( e )
        
        if t is None:
            t = et
        elif t is not et:
            raise ValueError( "Calling list_type on a list with at least two types ({0} and {1})".format( t, et ) )
    
    if t is None:
        raise ValueError( "Calling list_type on a list with no elements." )
    
    return t


def dict_to_list( v: dict ) -> List[str]:
    """
    Converts a string dictionary to a string list where each element is "x = y"
    """
    
    r = []
    
    for k, v in v.items():
        r.append( "{}={}".format( k, v ) )
    
    return r


def decomplex( item: Union[Optional[object], List[Optional[object]]] ) -> Optional[object]:
    """
    Converts items in arrays to single items
    """
    
    # noinspection PyBroadException
    try:
        if len( item ) == 1 and item[0] is not item:
            return decomplex( item[0] )
    except:
        pass
    
    return item


def as_sequence( contents: Union[T, List[T], Tuple[T]], elemental_none = False, cast = None, accept = (list, tuple) ) -> Union[List[T], Tuple[T]]:
    """
    The contents, as a tuple or list, if they are a tuple or list
    Or the sole item in a tuple
    
    Note: only `tuple` and `list` are handled, other `iterable`s are still placed in a tuple because we cannot know their nature.
    
    :param contents:        List contents
    :param elemental_none:  Doesn't treat `None` as an empty list
    :param cast:            Casts the result to this list type, if it isn't already 
    :param accept:          Types to accept
    :return: The result, as a list
    """
    if not elemental_none and contents is None:
        contents = tuple()
    elif not any( isinstance( contents, x ) for x in accept ):
        contents = contents,
    
    if cast is not None and not isinstance( contents, cast ):
        contents = cast( contents )
    
    return contents


def as_iter( contents: Union[T, List[T], Tuple[T]] ) -> Tuple[T]:
    """
    DEPRECATED
    """
    warnings.warn( "as_iter is deprecated, use `as_sequence`.", DeprecationWarning )
    return as_sequence( contents, elemental_none = True )


def as_tuple( contents: Union[T, List[T], Tuple[T]] ) -> Tuple[T]:
    """
    DEPRECATED
    """
    warnings.warn( "as_tuple is deprecated, use `as_sequence`.", DeprecationWarning )
    return as_sequence( contents, elemental_none = True, cast = tuple )


def as_list( contents: Union[T, List[T], Tuple[T]] ) -> List[T]:
    """
    DEPRECATED
    """
    warnings.warn( "as_list is deprecated, use `as_iter`.", DeprecationWarning )
    return as_sequence( contents, elemental_none = True, cast = list )


def backwards_range( count ):
    """
    A range, going backwards.
    """
    return range( count - 1, -1, -1 )


# !has test case
def create_index_lookup( source: Iterable[T] ) -> Dict[T, int]:
    """
    Creates a lookup table (`dict`) that allows the index of elements in `the_list` to be found.
    """
    result = { }
    result.update( (v, i) for i, v in enumerate( source ) )
    
    return result


def deinterleave_as_two( source: Iterable[T] ) -> Tuple[List[T], List[T]]:
    """
    Deinterleaves a source list, returns two new lists
    """
    dest_a = []
    dest_b = []
    iterator = iter( source )
    
    for a in iterator:
        dest_a.append( a )
        dest_b.append( next( iterator ) )
    
    return dest_a, dest_b


def deinterleave_as_iterator( source: Iterable[T] ) -> Iterator[Tuple[T, T]]:
    """
    Deinterleaves a source list, returns an iterator over tuples
    """
    iterator = iter( source )
    
    for a in iterator:
        yield a, next( iterator )


def deinterleave_as_list( source: Iterable[T] ) -> List[Tuple[T, T]]:
    """
    Deinterleaves a source list "A,B,A,B,...", returns a list of tuples "A, B"
    """
    dest_a = []
    iterator = iter( source )
    
    for a in iterator:
        dest_a.append( (a, next( iterator )) )
    
    return dest_a


def deinterleave_as_dict( source: Iterable[T] ) -> Dict[object, Optional[object]]:
    """
    Deinterleaves a source list "K,V,K,V,...", returns a dictionary "D" of "V = D[K]"
    """
    return dict( deinterleave_as_iterator( source ) )


def has_any( sequence: Iterable ) -> bool:
    """
    Returns if the sequence contains _any_ elements (i.e. non-zero length).
    """
    for _ in sequence:
        return True
    
    return False


def iterate_all( root, fn = None ) -> Iterator[Optional[object]]:
    """
    Iterates all items and descendants
    :param root: Where to start 
    :param fn: How to get the children 
    :return: Iterator over items and all descendants 
    """
    if fn is None:
        fn = lambda x: x
    
    for x in fn( root ):
        yield x
        yield from iterate_all( x, fn )


def ensure_capacity( array, index, value = None ):
    if len( array ) <= index:
        needed = index + 1 - len( array )
        array += [value] * needed


def index_of_first( array, elements ) -> Optional[int]:
    for e in elements:
        for i, x in enumerate( array ):
            if x == e:
                return i
    
    return None


class Indexer:
    """
    Provides a name to index and index to name lookup table.
    
    Note that `Indexer` has no indexer - be specific, use:
        `Indexer.name`
        `Indexer.index`
    """
    
    
    def __init__( self, iterator: Iterable[object] = None ):
        """
        CONSTRUCTOR
        Allows initialisation from existing entries
        """
        self.indexes = { }  # names to indices
        self.names = []  # indices to names
        
        if iterator is not None:
            for name in iterator:
                self.add( name )
    
    
    def add( self, name: object ):
        """
        Adds a new name with a new index.
        """
        index = self.indexes.get( name )
        
        if index is not None:
            return index
        
        index = len( self )
        
        self.indexes[name] = index
        self.names.append( name )
        
        return index
    
    
    def __len__( self ) -> int:
        """
        OVERRIDE
        Obtains the number of names
        """
        return len( self.names )
    
    
    def index( self, name: object ) -> int:
        """
        Obtains the index of the specified name.
        """
        return self.indexes[name]
    
    
    def name( self, index: int ):
        """
        Obtains the name at the specified index.
        """
        return self.names[index]


def first_or_none( array: Iterable[T], default = None ) -> Optional[T]:
    for item in array:
        return item
    
    return default


def first_or_error( array: Iterable[T] ) -> T:
    """
    Returns the first element of the array, raising a `KeyError` if the array is empty.
    """
    for item in array:
        return item
    
    raise KeyError( "Cannot extract the first element of the iterable because the iterable has no elements." )


def single_or_none( array: Iterable[T] ) -> Optional[T]:
    """
    Returns the first element of the array, returning `None` if the length is not `1`.
    """
    it = iter( array )
    
    try:
        first = next( it )
    except StopIteration:
        return None
    
    try:
        next( it )
        return None
    except StopIteration:
        return first


def single( array: Iterable[T] ) -> Optional[T]:
    """
    Returns the first element of the array, raising a `KeyError` if the length is not `1`.
    """
    return single_or_error( array )


def single_or_error( array: Iterable[T] ) -> Optional[T]:
    """
    Returns the first element of the array, raising a `KeyError` if the length is not `1`.
    """
    from mhelper import string_helper
    it = iter( array )
    
    try:
        first = next( it )
    except StopIteration:
        raise KeyError( "Cannot extract the single element of the iterable because the iterable has no elements." )
    
    try:
        next( it )
        raise KeyError( "Cannot extract the single element of the iterable because the iterable has multiple elements: {}".format( string_helper.format_array( array, limit = 10 ) ) )
    except StopIteration:
        return first


def lagged_iterate( sequence: Iterable[Optional[T]], head = False, tail = False ) -> Iterator[Tuple[Optional[T], Optional[T]]]:
    """
    Iterates over all adjacent pairs in the sequence. 
    
    :param sequence:        Sequence to iterate over `(0, 1, 2, 3, ..., n)` 
    :param head:            Include the head element `(None, 0)`. (off by default) 
    :param tail:            Include the tail element `(n, None)`. (off by default)
    :return:                The iteration: `(0,1), (1,2), (2,3), (...,...), (n-1,n)`
    
                                `head`  `tail`      `result when sequence = (1)`     `result when sequence = (1, 2, 3)`
                                False   False                                                   (1, 2), (2, 3)
                                True    False       (None, 1)                        (None, 1), (1, 2), (2, 3)  
                                True    True        (None, 1) (1, None)              (None, 1), (1, 2), (2, 3), (3, None)
                                False   True                  (1, None)                         (1, 2), (2, 3), (3, None)
                                
    """
    has_any = 0
    previous = None
    
    for current in sequence:
        if has_any:
            yield previous, current
        elif head:
            yield None, current
        
        has_any += 1
        previous = current
    
    if tail:
        yield previous, None


def triangular_comparison( sequence: List[T] ) -> Iterator[Tuple[T, T]]:
    """
    Order independent yielding of every element vs every other element.
    
    Pairs are only listed once, so if (A,B) is yielded, (B,A) is not.
    Self comparisons (A,A) are never listed.
    """
    for index, alpha in enumerate( sequence ):
        for beta in sequence[index + 1:]:
            yield alpha, beta


def square_comparison( sequence: List[T] ) -> Iterator[Tuple[T, T]]:
    """
    Order dependent yielding of every element vs every other element.
    
    Pairs are listed, so if (A,B) is yielded, so will (B,A).
    Self comparisons (A,A) are never listed.
    :param sequence: 
    :return: 
    """
    for index, alpha in enumerate( sequence ):
        for beta in sequence[:index]:
            yield alpha, beta
        
        for beta in sequence[index + 1:]:
            yield alpha, beta


class Single:
    def __init__( self, value = 0 ):
        self.value = value
    
    
    def next( self ):
        self.value += 1
        return self.value


def sort_two( a: T, b: U ) -> Tuple[T, U]:
    if a > b:
        return b, a
    else:
        return a, b


def sort_two_by_key( a: T, b: U, ka, kb ) -> Tuple[T, U]:
    if ka > kb:
        return b, a
    else:
        return a, b


def ordered_insert( list: List[T], item: T, key: Callable[[T], object] ):
    """
    Inserts the `item` into the `list` that has been sorted by `key`.
    """
    import bisect
    list.insert( bisect.bisect_left( [key( x ) for x in list], key( item ) ), item )


def average( list ) -> float:
    """
    Returns the mean average.
    """
    return sum( list ) / len( list )


def count( list ):
    try:
        return len( list )
    except:
        return sum( 1 for _ in list )


def when_last( i: Iterable[T] ) -> (T, bool):
    f = True
    l = None
    
    for x in i:
        if not f:
            yield l, False
        else:
            f = False
        
        l = x
    
    yield l, True


def when_first_or_last( iterable: Iterable[T] ) -> (T, bool, bool):
    is_first = True
    last_item = None
    has_yielded_first = True
    
    for item in iterable:
        if not is_first:
            yield last_item, has_yielded_first, False
            has_yielded_first = False
        else:
            is_first = False
        
        last_item = item
        
    if not is_first:
        yield last_item, has_yielded_first, True


def make_dict( **kwargs ):
    return kwargs


def find( i, p, default = NOT_PROVIDED ):
    for x in i:
        if p( x ):
            return x
    
    if default is not NOT_PROVIDED:
        return default
    else:
        raise ValueError( "No such value." )


def iter_distance_range( min: int, maxx: int, start: int ) -> Iterator[int]:
    yield start
    i = 1
    while True:
        if (start - i) >= min:
            yield start - i
        
        if (start + i) < maxx:
            yield start + i
        
        if (start - i) < min and (start + i) >= maxx:
            return
        
        i += 1


def distance_range( min: int, maxx: int, start: int ) -> List[int]:
    return list( iter_distance_range( min, maxx, start ) )


def add_to_listdict( dict_, key, value ):
    list_ = dict_.get( key )
    
    if list_ is None:
        list_ = []
        dict_[key] = list_
    
    list_.append( value )


def divide_workload( total_workload: int, num_workers: int, expand: bool = False ) -> List[Tuple[int, int]]:
    """
    Divides a workload of discrete objects between workers.
    
    :param total_workload:      The total workload to divide 
    :param num_workers:         The number of workers
    :param expand:              When true, `num_workers` specifies the maximum work a single worker can do.
                                The number of workers is calculated from this. 
    :return:                    A list of tuples, denoting the start and end of each worker's portion.
                                `len(result)` is thus the number of workers.
    """
    if expand:
        num_workers = int( 0.5 + total_workload / num_workers )
    
    r = []
    
    for index in range( num_workers ):
        r.append( get_workload( index, total_workload, num_workers ) )
    
    return r


def get_workload( index: int, total_workload: int, num_workers: int ) -> Tuple[int, int]:
    """
    Divides a workload of discrete objects between workers, and obtains the specified worker's portion.
    
    :param index:               The index 'i' of the worker to acquire the workload for. 
    :param total_workload:      The total workload 
    :param num_workers:         The number of workers 
    :return:                    A tuple denoting the start and end of the i'th worker's portion. 
    """
    worker_size = total_workload / num_workers
    
    start = index * worker_size
    
    if index == num_workers - 1:
        end = total_workload
    else:
        end = int( start + worker_size )
    
    return int( start ), end


def cross( l: Iterable[T] ) -> Iterator[Tuple[T, T]]:
    """
    Yields every item in the list v every other item in the list.
    Items are not self crossed, so 1, 2, 3 does not yield 1 v 1.
    Items are only crossed once, so 1, 2, 3 yields 1 v 2 but not 2 v 1.
    
    :param l:   The source iterator. This must yield the elements in the same order each time it is called.
                A special handler is in place for `set` however. 
    """
    if isinstance( l, set ):
        l = list( l )
    
    for a in l:
        for b in l:
            if b is a:
                break
            
            yield a, b



def pl( target, find = None ):
    if find is not None:
        find = find.lower()
    
    for index, member in enumerate( dir( target ) ):
        if find is not None and find not in member.lower():
            continue
        
        print( "    [{}] {}".format( index, member ) )


def ps( sequence ):
    for index, item in enumerate( sequence ):
        print( "    [{}] {}".format( index, item ) )



def coalesce( value, default_ ):
    """
    The null-coalescing operator, which is strangely missing from Python.
    """
    if value is None:
        return default_
    else:
        return value



"""
Functions for writing SVG images to disk.
"""
from enum import Enum

__author__ = "Martin Rusilowicz"


def _get_attributes( **kwargs ):
    r = []
    
    for k, v in kwargs.items():
        if v is not None:
            k = k.replace( "_", "-" )
            
            if isinstance( v, list ) or isinstance( v, tuple ):
                v = ",".join( str( x ) for x in v )
            elif isinstance( v, Enum ):
                v = v.name
                
                if v.startswith( "n_" ):
                    v = v[2:]
                
                v = v.replace( "_", "-" )
            
            r.append( '{}="{}"'.format( k, v ) )
    
    return " ".join( r )


class EFontWeight( Enum ):
    normal = 1
    bold = 2
    bolder = 3
    lighter = 4
    n_100 = 5
    n_200 = 6
    n_300 = 7
    n_400 = 8
    n_500 = 9
    n_600 = 10
    n_700 = 11
    n_800 = 12
    n_900 = 13
    inherit = 14


class EPriority:
    LINES = 1
    RECT = 2
    TEXT = 3


class ETextAnchor( Enum ):
    start = 1
    middle = 2
    end = 3


class EAlignmentBaseline( Enum ):
    auto = 1
    baseline = 2
    before_edge = 3
    text_before_edge = 4
    middle = 5
    central = 6
    after_edge = 7
    text_after_edge = 8
    ideographic = 9
    alphabetic = 10
    hanging = 11
    mathematical = 12
    inherit = 13


TColour = str
TMarker = str
TPos = float
TFraction = float
TDashArray = str
TPriority = int


class SvgWriter:
    """
    :attribute enable_html:     Output with HTML headers. Off by default.
    :attribute enable_autosize: Automatically expand width and height to accommodate new content. On by default.
    :attribute enable_kwargs:   Permit unknown arguments to the `add_` functions. Off by default.
    """
    
    
    def __init__( self ):
        self.__defs = []
        self.__data = []
        
        self.enable_html = False
        self.enable_autosize = True
        self.enable_kwargs = False
        self.width = 128
        self.height = 128
        
        self.__defs.append( '<marker id="markerCircle" markerWidth="8" markerHeight="8" refX="5" refY="5"><circle cx="5" cy="5" r="3" style="stroke: none; fill:#00C000;"/></marker><marker id="markerArrow" markerWidth="13" markerHeight="13" refX="2" refY="6" orient="auto"><path d="M2,2 L2,11 L10,6 L2,2" style="fill: #00C000;" /></marker>' )
    
    
    def _get_header_and_footer( self ):
        if self.enable_html:
            header = '<html><body><svg width="{}" height="{}">'
            footer = '</svg></body></html>'
        else:
            header = '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="{}" height="{}">'
            footer = '</svg></body></html>'
        
        defs = '<defs>' + "\n".join( self.__defs ) + '</defs>'
        header = header.format( self.width, self.height )
        header += defs
        
        return header, footer
    
    
    def to_string( self ):
        r = []
        header, footer = self._get_header_and_footer()
        r.append( header )
        r.extend( x[0] for x in sorted( self.__data, key = self._get_order ) )
        r.append( footer )
        
        return "\n".join( r )
    
    
    def _get_order( self, line ):
        return line[1] * 10000 + line[2]
    
    
    def _append( self, z, content ):
        self.__data.append( (content, z, len( self.__data )) )
    
    
    def _check_kwargs( self, kwargs ):
        if not self.enable_kwargs and kwargs:
            raise ValueError( "Unknown arguments passed. Did you mean to pass these parameters? If you did set `enable_kwargs` to `True` to enable unknown parameters. kwargs = {}".format( kwargs ) )
    
    
    def add_connector( self,
                       x1: TPos,
                       y1: TPos,
                       w1: TPos,
                       h1: TPos,
                       x2: TPos,
                       y2: TPos,
                       w2: TPos,
                       h2: TPos,
                       **kwargs ):
        return self.add_line( x1 + w1 / 2, y1 + h1 / 2, x2 + w2 / 2, y2 + h2 / 2, **kwargs )
    
    
    def add_line( self,
                  x1: TPos,
                  y1: TPos,
                  x2: TPos,
                  y2: TPos,
                  z: TPriority = EPriority.LINES,
                  *,
                  stroke: TColour = None,
                  stroke_width: TPos = None,
                  marker_end: TMarker = None,
                  **kwargs ):
        self._check_kwargs( kwargs )
        self.ensure_size( x1, y1 )
        self.ensure_size( x2, y2 )
        self._append( z, '<line {} />'.format( _get_attributes( x1 = x1, y1 = y1, x2 = x2, y2 = y2, stroke = stroke, stroke_width = stroke_width, marker_end = marker_end, **kwargs ) ) )
    
    
    def add_rect( self,
                  x: TPos,
                  y: TPos,
                  w: TPos,
                  h: TPos,
                  z: TPriority = EPriority.RECT,
                  *,
                  rx = TPos,
                  ry = TPos,
                  fill = TColour,
                  stroke = TColour,
                  stroke_width = TPos,
                  stroke_dasharray = TDashArray,
                  fill_opacity = TFraction,
                  stroke_opacity = TFraction,
                  **kwargs ):
        self._check_kwargs( kwargs )
        self.ensure_size( x + w, y + w )
        self._append( z, '<rect {} />'.format( _get_attributes( x = x, y = y, width = w, height = h, rx = rx, ry = ry, fill = fill, stroke = stroke, stroke_width = stroke_width, stroke_dasharray = stroke_dasharray, fill_opacity = fill_opacity, stroke_opacity = stroke_opacity, **kwargs ) ) )
    
    
    def centre_text( self,
                     x: TPos,
                     y: TPos,
                     w: TPos,
                     h: TPos,
                     text: str,
                     **kwargs ):
        self.add_text( x + w / 2, y + h / 2, text, alignment_baseline = EAlignmentBaseline.middle, text_anchor = ETextAnchor.middle, **kwargs )
    
    
    def add_text( self,
                  x: TPos,
                  y: TPos,
                  text: str,
                  z: TPriority = EPriority.TEXT,
                  *,
                  font_family: str = None,
                  font_size: TPos = None,
                  font_weight: EFontWeight = None,
                  fill: TColour = None,
                  alignment_baseline: EAlignmentBaseline = None,
                  text_anchor: ETextAnchor = None,
                  **kwargs ):
        self._check_kwargs( kwargs )
        self.ensure_size( x, y )
        self._append( z, '<text {}>{}</text>'.format( _get_attributes( x = x, y = y, font_family = font_family, font_size = font_size, font_weight = font_weight, fill = fill, alignment_baseline = alignment_baseline, text_anchor = text_anchor, **kwargs ), text ) )
    
    
    def ensure_size( self, x, y ):
        if not self.enable_autosize:
            return
        
        if self.width < x:
            self.width = x
        
        if self.height < y:
            self.height = y



from typing import Tuple, Iterable


class Serialisable:
    """
    Provides overridable methods for dealing with serialisation easier.
    
        ```
        _SERIALISABLE_state
        ``` 
        
    Unlike `__getstate__` __setstate__`:
        * The state only needs to be specified once.
        * The non-serialised fields are specified rather than the serialisable ones.
    """
    
    
    def _SERIALISABLE_state( self ) -> Iterable[Tuple[type, str, object]]:
        """
        Gets the serialisation state.
        
        :return: The derived class should return (typically through a sequence of `yield` statements)
                 an iterable over tuples describing the non-serialisable objects.
            Iterable:
                Tuple:
                    type:   The type of the class on which the field is present
                    str:    The name of the field
                    object: The default value of the field when the object is deserialized
        """
        return []
    
    
    def __get_state( self, test ):
        d = { }
        
        for class_, field, value in self._SERIALISABLE_state():
            if field.startswith( "__" ) and not field.endswith( "__" ):
                field = "_{}{}".format( class_.__name__, field )
            
            if test and field not in self.__dict__:
                raise ValueError( "No such field as `{}`.".format( field ) )
            
            d[field] = value
        
        return d
    
    
    def __getstate__( self ):
        result = dict( self.__dict__ )
        
        for x in self.__get_state( True ).keys():
            del result[x]
        
        return result
    
    
    def __setstate__( self, state ):
        for k, v in state.items():
            setattr( self, k, v )
        
        for k, v in self.__get_state( False ).items():
            setattr( self, k, v )



"""
Load and save functions
"""
import sys
import warnings
from typing import Any, Optional, TypeVar, Union, cast


T = TypeVar( "T" )

CAUSE_ERROR = object  # Sentinel object for load_* functions


def load_npy( file_name ):
    import numpy
    result = numpy.load( file_name, allow_pickle = False, fix_imports = False )
    
    if file_name.upper().endswith( ".NPZ" ):
        result = result["data"]
    
    return result


def load_npz( file_name ):
    import numpy
    result = numpy.load( file_name, allow_pickle = False, fix_imports = False )
    result = result["data"]
    
    return result


def save_bitarray( file_name: str, value ) -> None:
    """
    Saves a bit array
    
    :type value: bitarray
    
    :param file_name: File name
    :param value: Value to save
    """
    from bitarray import bitarray
    from mhelper.exception_helper import SubprocessError
    from mhelper import exception_helper, file_helper
    
    exception_helper.assert_instance( "save_bitarray::value", value, bitarray )
    assert isinstance( value, bitarray )
    
    try:
        with open( file_name, "wb" ) as file_out:
            value.tofile( file_out )
    except TypeError:
        # Stupid "open file expected" error on OSX (due to bug writing large files - fallback to manual implementation)
        _save_bytes_manually( file_name, value.tobytes() )
    except Exception as ex:
        raise SubprocessError( "Failed to write bitarray of length {} to «{}».".format( len( value ), file_name ) ) from ex
    
    size = float( file_helper.file_size( file_name ) )
    expected = len( value )
    
    if size < expected / 8.0:
        raise ValueError( "Saved file is shorter ({} bytes or {} bits) than the originating bitarray ({} bytes or {} bits).".format( size, size * 8, expected / 8, expected ) )


def _save_bytes_manually( file_name: str, value: bytes ):
    """
    Fallback function used by `save_bitarray`.
    """
    warnings.warn( "Save bitarray failed. This is probably due to an error in your `bitarray` library. The file will be saved incrementally but this will take longer.", UserWarning )
    BATCH_SIZE = 100000
    cursor = 0
    length = len( value )
    
    with open( file_name, "wb" ) as file_out:  # note that writing large arrays on OSX is probably the problem, we can't just dump the bytes
        while cursor < length:
            next = min( cursor + BATCH_SIZE, length )
            slice = value[cursor:next]
            file_out.write( slice )
            cursor = next


def _read_bytes_manually( file_name: str ) -> bytes:
    BATCH_SIZE = 100000
    b = bytearray()
    
    with open( file_name, "rb" ) as file_in:
        while True:
            buf = file_in.read( BATCH_SIZE )
            
            if len( buf ) == 0:
                break
            
            b.extend( buf )
    
    return bytes( b )


def load_bitarray( file_name: str ):
    """
    Loads a bitarray
    :param file_name: File to read 
    :return: bitarray. Note this may be padded with missing bits. 
    """
    from bitarray import bitarray
    
    result = bitarray()
    
    try:
        with open( file_name, 'rb' ) as file_in:
            result.fromfile( file_in )
    except SystemError:  # OSX error
        result = bitarray()
        result.frombytes( _read_bytes_manually( file_name ) )
        assert len( result ) != 0
    
    return result


def save_npy( file_name, value ):
    import numpy
    
    numpy.save( file_name, value, allow_pickle = False, fix_imports = False )


def save_npz( file_name, value ):
    import numpy
    
    numpy.savez_compressed( file_name, data = value )


def load_binary( file_name: str, default: Optional[object] = CAUSE_ERROR ):
    """
    Loads a binary file
    
    :param file_name: Filename to load  
    :param default:     Whether to default the values (if not specified or `cause_error` an exception is raised) 
    :return: Loaded file, or `default` (if specified) 
    """
    try:
        import pickle
        
        with open( file_name, "rb" ) as file:
            return pickle.load( file, fix_imports = False )
    except Exception:
        if default is CAUSE_ERROR:
            raise
        
        return default
    except BaseException as ex:
        warnings.warn( "Loading the binary file «{}» caused a «{}» exception.".format( file_name, type( ex ).__name__ ), UserWarning )
        raise


def save_binary( file_name, value ):
    import pickle
    
    with open( file_name, "wb" ) as file:
        pickle.dump( value, file, fix_imports = False )


def load_json( file_name, keys = True ):
    from mhelper.file_helper import read_all_text
    import jsonpickle
    
    text = read_all_text( file_name )
    
    return jsonpickle.decode( text, keys = keys )


def save_json( file_name, value, keys = True ):
    from mhelper.file_helper import write_all_text
    import jsonpickle
    
    write_all_text( file_name, jsonpickle.encode( value, keys = keys ) )


def default_values( target: T, default: Optional[Union[T, type]] = None ) -> T:
    if default is None:
        if target is None:
            raise ValueError( "Cannot set the defaults for the value because both the value and the defaults are `None`, so neither can be inferred." )
        
        default = type( target )
    
    if isinstance( default, type ):
        default = default()
    
    if target is None:
        return default
    
    if isinstance( target, list ):
        return cast( T, target )
    
    # TODO: PUT THIS BACK IN! Removed because for some reason we get a different class when deserialising from the IDE, but this is a problem!
    # if type(target) is not type(default):
    #     raise ValueError("Attempting to set the defaults for the value «{}», of type «{}», but the value provided, «{}», of type «{}», is not compatible with this.".format(target, type(target), default, type(default)))
    
    for k, v in default.__dict__.items():
        if k.startswith( "_" ):
            continue
        
        if k not in target.__dict__:
            target.__dict__[k] = v
    
    to_delete = []
    
    for k in target.__dict__.keys():
        if k not in default.__dict__:
            to_delete.append( k )
    
    for k in to_delete:
        del target.__dict__[k]
    
    return target


class StdOutWriter:
    """
    Handles writing to STDOUT instead of a file.
    """
    
    
    def __init__( self, write_fn = None ) -> None:
        if write_fn is None:
            write_fn = sys.stdout.write
        
        self.lines = []
        self.closed = False
        self.write_fn = write_fn
    
    
    def __enter__( self ):
        return self
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        self.close()
    
    
    def write( self, text ) -> None:
        if self.closed:
            raise ValueError( "Virtual stream has already been closed." )
        
        self.lines.append( text )
    
    
    def close( self ) -> None:
        if self.closed:
            return
        
        for line in self.lines:
            self.write_fn( line )
        
        self.closed = True
    
    
    def __repr__( self ):
        return "StdOutWriter"


class NowhereWriter:
    """
    Handles writing to nothing instead of a file.
    """
    
    
    def __init__( self ) -> None:
        pass
    
    
    def __enter__( self ):
        return self
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        self.close()
    
    
    def write( self, text ) -> None:
        pass
    
    
    def close( self ) -> None:
        pass
    
    
    def __repr__( self ):
        return "NowhereWriter"


def open_write( file_name: str, map = None ) -> Any:
    """
    Opens a file for writing, also accepts `stdout` and `null` as the file name.
    """
    if file_name is None:
        file_name_lower = None
    else:
        file_name_lower = file_name.lower()
    
    mapped = map.get( file_name_lower )
    
    if mapped is not None:
        return StdOutWriter( mapped )
    
    if file_name is None or file_name_lower in ("", "stdout"):
        return StdOutWriter()
    elif file_name_lower == "null":
        return NowhereWriter()
    else:
        return open( file_name, "w" )



import subprocess
from subprocess import Popen
from queue import Queue, Empty
from threading import Event, Thread
from typing import Callable, Sequence, IO

from mhelper.exception_helper import SubprocessError


DOnOutput = Callable[[str], None]


def __enqueue_stream( out: IO, queue: Queue, event ) -> None:
    for line in out:
        line = line.decode()
        if line.endswith( "\n" ):
            line = line[:-1]
        queue.put( line )
        event.set()
    
    out.close()


def __wait_exit( process: subprocess.Popen, event: Event ) -> None:
    process.wait()
    event.set()


def async_run( cmds: Sequence[str], on_stdout: DOnOutput, on_stderr: DOnOutput, check: bool = False ) -> int:
    """
    Runs a process asynchronously.
    This function blocks until the process completes, but output is diverted to the provided functions.
    
    Lines received from the standard streams are sent to `on_stdout` and `on_stderr`.
    Any trailing \n are removed.
    Threading is handled, hence these functions are called in the same thread as the one that calls `async_run`. 
     
    :param cmds:                Command sequence to run. See `Popen`.
    :param on_stdout:           Lines received from standard output are sent here. 
    :param on_stderr:           Lines received from standard error are sent here.
    :param check:               When True, a non-zero error code raises an Exception.
    :return:                    The return code of the program.
    :except SubprocessError:    Non-zero error code and `check` was passed as true.
    """
    process = Popen( cmds, stdout = subprocess.PIPE, stderr = subprocess.PIPE, stdin = subprocess.PIPE )
    process.stdin.close()
    
    stdout_queue = Queue()
    stderr_queue = Queue()
    event = Event()
    
    thread_1 = Thread( target = __enqueue_stream, args = (process.stdout, stdout_queue, event) )
    thread_1.name = "async_run.stdout_thread(" + " ".join( cmds ) + ")"
    thread_1.daemon = True
    thread_1.start()
    
    thread_2 = Thread( target = __enqueue_stream, args = (process.stderr, stderr_queue, event) )
    thread_2.name = "async_run.stderr_thread(" + " ".join( cmds ) + ")"
    thread_2.daemon = True
    thread_2.start()
    
    thread_3 = Thread( target = __wait_exit, args = (process, event) )
    thread_2.name = "async_run.exit_thread(" + " ".join( cmds ) + ")"
    thread_3.daemon = True
    thread_3.start()
    
    while True:
        event.wait()
        event.clear()
        
        while True:
            used = False
            
            try:
                line = stdout_queue.get_nowait()
            except Empty:
                pass
            else:
                on_stdout( line )
                used = True
            
            try:
                line = stderr_queue.get_nowait()
            except Empty:
                pass
            else:
                on_stderr( line )
                used = True
            
            if not used:
                break
        
        if process.returncode is not None:
            if check and process.returncode:
                raise SubprocessError( "The command «{}» exited with error code «{}». If available, checking the output may provide more details.".format( " ".join( '"{}"'.format( x ) for x in cmds ), process.returncode ) )
            
            return process.returncode



import re
from Bio import Phylo
from Bio.Phylo.BaseTree import Tree
from io import StringIO
from typing import Iterator, Tuple


_REFX = re.compile( "([0-9.]+):[0-9.]+" )


def convert_file( in_filename, out_filename, in_format, out_format ):
    from Bio import AlignIO
    
    with open( in_filename, "rU" ) as input_handle:
        with open( out_filename, "w" ) as output_handle:
            alignments = AlignIO.parse( input_handle, in_format )
            AlignIO.write( alignments, output_handle, out_format )


def parse_fasta( *, text = None, file = None ) -> Iterator[Tuple[str, str]]:
    """
    Parses a FASTA file.
    Accepts multi-line sequences.
    Accepts ';' comments in the file.
    
    Nb. BioPython's SeqIO.parse doesn't handle comments for FASTA.
    
    :param text:    FASTA text 
    :param file:    Path to FASTA file 
    :return:        Tuples of sequence names and sites 
    """
    from mhelper import file_helper
    
    if file is not None:
        if text is not None:
            raise ValueError( "Cannot specify both `file` and `text` arguments to `parse_fasta`." )
        
        text = file_helper.read_all_text( file )
    elif text is None:
        raise ValueError( "Must specify either `file` or `text` arguments when calling `parse_fasta`." )
    
    heading = None
    sequence = []
    
    for line in text.split( "\n" ):
        line = line.strip()
        
        if line.startswith( ">" ):
            if heading is not None:
                yield heading, "".join( sequence )
            
            heading = line[1:]
            sequence = []
        elif not line.startswith( ";" ):
            sequence.append( line )
    
    if heading is not None:
        yield heading, "".join( sequence )


def biotree_to_newick( tree: Tree ) -> str:
    handle = StringIO()
    Phylo.write( [tree], handle, "newick" )
    result = handle.getvalue()
    
    # Work around stupid BioPython bug
    # https://github.com/biopython/biopython/issues/1315
    # TODO: Remove this fix when the issue is fixed
    result = _REFX.sub( "\\1", result )
    
    return result


def newick_to_biotree( newick ) -> Tree:
    handle = StringIO( newick )
    return Phylo.read( handle, "newick" )



class BatchList:
    def __init__( self, data, batch_size ):
        self.batch = batch_size
        self.data = list( data )
    
    
    def take( self ):
        result = self.data[0:self.batch]
        del self.data[0:self.batch]
        return result
    
    
    def __bool__( self ):
        return bool( self.data )
    
    
    def __len__( self ):
        return len( self.data )
    
    
    def __iter__( self ):
        while self:
            yield self.take()


def divide_workload( index, count, quantity ):
    import warnings
    warnings.warn( "deprecated - array_helper.get_workload", DeprecationWarning )
    from mhelper import array_helper
    return array_helper.get_workload( index, quantity, count )



from typing import Optional, Callable


class ManagedWith:
    """
    Wraps a (presumably pooled) object.
    
    Usage:

        ```
        with ManagedWith(...):
            ...
        ```
        
    Example (database provider):
    
        def get_database():
            return ManagedWith(pool.pop(), pool.push)
        
    Example (from `Plugin`):
        
        ```
        with get_database() as db:
            db.execute("RETURN 1")
        ```
    """
    
    
    def YOURE_USING_THE_WRONG_THING( self ):
        """
        Message to an IDE user telling them they're using the wrong thing! (should be using `with`)
        """
        pass
    
    
    def __init__( self,
                  target: Optional[ object ] = None,
                  on_exit: Callable[[Optional[ object]], None] = None,
                  on_enter: Callable[[Optional[ object]], None] = None,
                  on_get_target: Callable[[ ], object] = None ):
        """
        :param target: Object to manage 
        :param on_exit: What to do when the caller has finished with the target 
        """
        self.__target = target
        self.__on_exit = on_exit
        self.__on_enter = on_enter
        self.__on_get_target = on_get_target
        
        if self.__on_get_target is not None and self.__target is not None:
            raise ValueError( "Cannot specify both 'on_get_target' and 'target' parameters." )
    
    
    def __enter__( self ):
        if self.__on_get_target is not None:
            self.__target = self.__on_get_target()
        
        if self.__on_enter is not None:
            self.__on_enter( self.__target )
        
        return self.__target
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        if self.__on_exit is not None:
            self.__on_exit( self.__target )
        
        if self.__on_get_target is not None:
            self.__target = None



from mhelper import array_helper
from mhelper import batch_lists
from mhelper import bio_helper
from mhelper import comment_helper
from mhelper import component_helper
from mhelper import disposal_helper
from mhelper import event_helper
from mhelper import exception_helper
from mhelper import file_helper
from mhelper import generics_helper
from mhelper import io_helper
from mhelper import log_helper
from mhelper import maths_helper
from mhelper import proxy_helper
from mhelper import string_helper
from mhelper import string_parser
from mhelper import svg_helper
from mhelper import reflection_helper
from mhelper import ansi_helper
from mhelper import misc_helper
from mhelper import async_helper

from mhelper.ansi_helper import AnsiStr
from mhelper.array_helper import Indexer
from mhelper.batch_lists import BatchList
from mhelper.comment_helper import abstract, ignore, override, overrides, protected, sealed, virtual
from mhelper.component_helper import ComponentFinder
from mhelper.disposal_helper import ManagedWith
from mhelper.exception_helper import ImplementationError, LogicError, NotFoundError, NotSupportedError, SwitchError, SubprocessError
from mhelper.generics_helper import ByRef, GenericString, GenericStringMeta, MAnnotation, MAnnotation, MAnnotationArgs, MAnnotationFactory, MGeneric, MGenericMeta, NonGenericString
from mhelper.log_helper import Logger, LOGGERS, DEFAULT_LOGGER
from mhelper.property_helper import itemproperty
from mhelper.proxy_helper import PropertySetInfo, SimpleProxy
from mhelper.qt_resource_objects import ResourceIcon
from mhelper.reflection_helper import AnnotationInspector, TTristate
from mhelper.serialisable import Serialisable
from mhelper.special_types import Dirname, EFileMode, FileNameAnnotation, Filename, HReadonly, MEnum, MFlags, MOptional, MUnion, NOT_PROVIDED, Password
from mhelper.string_helper import FindError
from mhelper.svg_helper import SvgWriter



__author__ = "Martin Rusilowicz"
__version__ = "1.0.1.41"



"""
Special ANSI codes.
"""

# Special
RESET = "\033[0m"

# Styles
BOLD = "\033[1m"
DIM = "\033[2m"
ITALIC = "\033[3m"
UNDERLINE = "\033[4m"
BLINK = "\033[5m"
INVERSE = "\033[7m"
NORMAL = "\033[22m"

# Codes
CODE_FORE_BLACK = 30
CODE_FORE_RED = 31
CODE_FORE_GREEN = 32
CODE_FORE_YELLOW = 33
CODE_FORE_BLUE = 34
CODE_FORE_MAGENTA = 35
CODE_FORE_CYAN = 36
CODE_FORE_WHITE = 37
CODE_FORE_RESET = 39
CODE_FORE_LIGHT_BLACK = 90
CODE_FORE_LIGHT_RED = 91
CODE_FORE_LIGHT_GREEN = 92
CODE_FORE_LIGHT_YELLOW = 93
CODE_FORE_LIGHT_BLUE = 94
CODE_FORE_LIGHT_MAGENTA = 95
CODE_FORE_LIGHT_CYAN = 96
CODE_FORE_LIGHT_WHITE = 97
CODE_BACK_BLACK = 40
CODE_BACK_RED = 41
CODE_BACK_GREEN = 42
CODE_BACK_YELLOW = 43
CODE_BACK_BLUE = 44
CODE_BACK_MAGENTA = 45
CODE_BACK_CYAN = 46
CODE_BACK_WHITE = 47
CODE_BACK_RESET = 49
CODE_BACK_LIGHT_BLACK = 100
CODE_BACK_LIGHT_RED = 101
CODE_BACK_LIGHT_GREEN = 102
CODE_BACK_LIGHT_YELLOW = 103
CODE_BACK_LIGHT_BLUE = 104
CODE_BACK_LIGHT_MAGENTA = 105
CODE_BACK_LIGHT_CYAN = 106
CODE_BACK_LIGHT_WHITE = 107
CODE_STYLE_RESET_ALL = 0
CODE_STYLE_BRIGHT = 1
CODE_STYLE_DIM = 2
CODE_STYLE_NORMAL = 22

# Internal codes
CODE_INTERNAL_DEFAULT = -1
CODE_INTERNAL_QUOTE_START = -2
CODE_INTERNAL_QUOTE_END = -3
CODE_INTERNAL_TABLE_START = -4
CODE_INTERNAL_TABLE_END = -5

# Foreground
FORE_BLACK = "\033[30m"
FORE_RED = "\033[31m"
FORE_GREEN = "\033[32m"
FORE_YELLOW = "\033[33m"
FORE_BLUE = "\033[34m"
FORE_MAGENTA = "\033[35m"
FORE_CYAN = "\033[36m"
FORE_WHITE = "\033[37m"
FORE_RESET = "\033[39m"
FORE_BRIGHT_BLACK = "\033[90m"
FORE_BRIGHT_RED = "\033[91m"
FORE_BRIGHT_GREEN = "\033[92m"
FORE_BRIGHT_YELLOW = "\033[93m"
FORE_BRIGHT_BLUE = "\033[94m"
FORE_BRIGHT_MAGENTA = "\033[95m"
FORE_BRIGHT_CYAN = "\033[96m"
FORE_BRIGHT_WHITE = "\033[97m"

# Background
BACK_BLACK = "\033[40m"
BACK_RED = "\033[41m"
BACK_GREEN = "\033[42m"
BACK_YELLOW = "\033[43m"
BACK_BLUE = "\033[44m"
BACK_MAGENTA = "\033[45m"
BACK_CYAN = "\033[46m"
BACK_WHITE = "\033[47m"
BACK_RESET = "\033[49m"
BACK_LIGHT_BLACK = "\033[100m"
BACK_LIGHT_RED = "\033[101m"
BACK_LIGHT_GREEN = "\033[102m"
BACK_LIGHT_YELLOW = "\033[103m"
BACK_LIGHT_BLUE = "\033[104m"
BACK_LIGHT_MAGENTA = "\033[105m"
BACK_LIGHT_CYAN = "\033[106m"
BACK_LIGHT_WHITE = "\033[107m"

# Do not use
BACK_BRIGHT_BLACK = "\033[100m"
BACK_BRIGHT_RED = "\033[101m"
BACK_BRIGHT_GREEN = "\033[102m"
BACK_BRIGHT_YELLOW = "\033[103m"
BACK_BRIGHT_BLUE = "\033[104m"
BACK_BRIGHT_MAGENTA = "\033[105m"
BACK_BRIGHT_CYAN = "\033[106m"
BACK_BRIGHT_WHITE = "\033[107m"

# 24-bit colours
def fore( r, g, b ):
    return "\033[38;2;" + str( r ) + ";" + str( g ) + ";" + str( b ) + "m"

def back( r, g, b ):
    return "\033[48;2;" + str( r ) + ";" + str( g ) + ";" + str( b ) + "m"

# Text
FR = FORE_RED
FG = FORE_GREEN
FB = FORE_BLUE
FC = FORE_CYAN
FY = FORE_YELLOW
FM = FORE_MAGENTA
FW = FORE_WHITE
FK = FORE_BLACK

FBR = FORE_BRIGHT_RED
FBG = FORE_BRIGHT_GREEN
FBB = FORE_BRIGHT_BLUE
FBC = FORE_BRIGHT_CYAN
FBY = FORE_BRIGHT_YELLOW
FBM = FORE_BRIGHT_MAGENTA
FBW = FORE_BRIGHT_WHITE
FBK = FORE_BRIGHT_BLACK

#Background
BR = BACK_RED
BG = BACK_GREEN
BB = BACK_BLUE
BC = BACK_CYAN
BY = BACK_YELLOW
BM = BACK_MAGENTA
BW = BACK_WHITE
BK = BACK_BLACK

BBR = BACK_LIGHT_RED
BBG = BACK_LIGHT_GREEN
BBB = BACK_LIGHT_BLUE
BBC = BACK_LIGHT_CYAN
BBY = BACK_LIGHT_YELLOW
BBM = BACK_LIGHT_MAGENTA
BBW = BACK_LIGHT_WHITE
BBK = BACK_LIGHT_BLACK

# Styles
SB = BOLD
SD = DIM
SI = ITALIC
SU = UNDERLINE
SE = INVERSE




import unittest
from typing import Optional, Union

import jsonpickle

from mhelper import BatchList, array_helper, batch_lists, string_helper, AnnotationInspector, ansi_helper, ansi, Serialisable, io_helper
from mhelper.property_helper import itemproperty


class MyClassSerialisable( Serialisable ):
    def __init__( self ):
        super().__init__()
        self.a = 1
        self.b = 2
        self.c = 3
        self.__d = 4
        self.__e = 5
        self.__f = 6
    
    
    def _SERIALISABLE_state( self ):
        yield MyClassSerialisable, "c", 3.1
        yield MyClassSerialisable, "__f", 6.1
    
    
    def set_values( self ):
        self.a = -1
        self.b = -2
        self.c = -3
        self.__d = -4
        self.__e = -5
        self.__f = -6
    
    
    def __str__( self ):
        return "{} {} {} {} {} {}".format( self.a, self.b, self.c, self.__d, self.__e, self.__f )


class serialisable_helper_tests( unittest.TestCase ):
    def test_functionality( self ):
        original = MyClassSerialisable()
        original.set_values()
        
        encoded = jsonpickle.encode( original )
        decoded = jsonpickle.decode( encoded )
        
        self.assertEqual( str( original ), "-1 -2 -3 -4 -5 -6" )
        self.assertEqual( str( decoded ), "-1 -2 3.1 -4 -5 6.1" )


class property_helper_tests( unittest.TestCase ):
    def test_decorator( self ):
        class temporary:
            def __init__( self ):
                self.items = { }
            
            
            @itemproperty
            def blah( self, key ):
                return self.items[key]
            
            
            @blah.itemsetter
            def blah( self, key, value ):
                self.items[key] = value
        
        
        a = temporary()
        
        a.blah[5] = "555"
        a.blah[8] = "888"
        self.assertEqual( a.blah[5], "555" )
        self.assertEqual( a.blah[8], "888" )


class ansi_helper_tests( unittest.TestCase ):
    def test_functions( self ):
        # TODO: These tests should check the output automatically
        
        my_text = ansi.FORE_RED + "RED" + ansi.FORE_GREEN + "GREEN" + ansi.FORE_BLUE + "BLUE" + ansi.RESET
        my_text_2 = (my_text + " ") * 20
        
        for width in (5, 10, 15):
            print( "            [" + "." * width + "]" )
            print( "fix_width = [{}]".format( ansi_helper.fix_width( my_text, width ) ) + ansi.RESET )
            print( "fix_width = [{}]".format( ansi_helper.fix_width( my_text, width, "~", -1 ) ) + ansi.RESET )
            print( "fix_width = [{}]".format( ansi_helper.fix_width( my_text, width, "~", 0 ) ) + ansi.RESET )
            print( "fix_width = [{}]".format( ansi_helper.fix_width( my_text, width, "~", 1 ) ) + ansi.RESET )
            print( "ljust     = [{}]".format( ansi_helper.ljust( my_text, width, "~" ) ) + ansi.RESET )
            print( "cjust     = [{}]".format( ansi_helper.cjust( my_text, width, "~" ) ) + ansi.RESET )
            print( "rjust     = [{}]".format( ansi_helper.rjust( my_text, width, "~" ) ) + ansi.RESET )
        
        print( "length    = [{}]".format( ansi_helper.length( my_text ) ) + ansi.RESET )
        print( "without   = [{}]".format( ansi_helper.without_ansi( my_text ) ) + ansi.RESET )
        print( "wrap20,80 = [{}]".format( "\n".join( ansi_helper.wrap( my_text_2, 80 ) ) ) + ansi.RESET )
        print( "add2      = [{}]".format( my_text + my_text ) + ansi.RESET )
        print( "multiply3 = [{}]".format( my_text * 3 ) + ansi.RESET )


class string_helper_tests( unittest.TestCase ):
    def test_highlight_quotes( self ):
        self.assertEqual( string_helper.highlight_quotes( "hello 'world' hello's", "'", "'", "<", ">" ), "hello <world> hello's" )
        self.assertEqual( string_helper.highlight_quotes( "alpha 'beta' gamma 'delta'", "'", "'", "<", ">" ), "alpha <beta> gamma <delta>" )
    
    
    def test_fix_indents( self ):
        text_a = """alpha\n\n        beta\n            gamma\n                delta\n        epsilon\n    """
        text_b = """alpha\n\nbeta\n    gamma\n        delta\nepsilon\n"""
        self.assertEqual( string_helper.fix_indents( text_a ), text_b )


class array_helper_tests( unittest.TestCase ):
    def test__create_index_lookup( self ):
        """
        TEST: create_index_lookup
        """
        
        my_list = "a", "b", "c", "d", "e"
        
        lookup = array_helper.create_index_lookup( my_list )
        
        # Sanity check
        for i, v in enumerate( my_list ):
            self.assertEqual( i, lookup[v] )


class batch_lists_test( unittest.TestCase ):
    def test__BatchList__take( self ):
        b = BatchList( (1, 2, 3, 4, 5, 6, 7, 8, 9, 10), 3 )
        
        self.assertEqual( b.take(), [1, 2, 3] )
        self.assertEqual( b.take(), [4, 5, 6] )
        self.assertEqual( b.take(), [7, 8, 9] )
        self.assertEqual( b.take(), [10, ] )
    
    
    def test__divide_workload( self ):
        self.assertEqual( batch_lists.divide_workload( 0, 5, 51 ), (0, 10) )
        self.assertEqual( batch_lists.divide_workload( 1, 5, 51 ), (10, 20) )
        self.assertEqual( batch_lists.divide_workload( 2, 5, 51 ), (20, 30) )
        self.assertEqual( batch_lists.divide_workload( 3, 5, 51 ), (30, 40) )
        self.assertEqual( batch_lists.divide_workload( 4, 5, 51 ), (40, 51) )
    
    
    def test__divide_workload_2( self ):
        last_b = None
        for i in range( 1000 ):
            a, b = batch_lists.divide_workload( i, 1000, 1600 )
            
            self.assertGreater( b, a )
            
            if last_b is not None:
                self.assertEqual( a, last_b )
            
            last_b = b


class reflection_helper_tests( unittest.TestCase ):
    def test__annotations( self ):
        # noinspection PyClassHasNoInit
        class UPPER:
            pass
        
        
        # noinspection PyClassHasNoInit
        class MIDDLE( UPPER ):
            pass
        
        
        # noinspection PyClassHasNoInit
        class LOWER( MIDDLE ):
            pass
        
        
        # noinspection PyClassHasNoInit
        class DIFFERENT:
            pass
        
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_indirectly_below( UPPER ) )
        self.assertTrue( AnnotationInspector( Optional[MIDDLE] ).is_indirectly_below( UPPER ) )
        self.assertTrue( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_below( UPPER ) )
        self.assertFalse( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_below( LOWER ) )
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_directly_below( UPPER ) )
        self.assertFalse( AnnotationInspector( Optional[MIDDLE] ).is_directly_below( UPPER ) )
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_indirectly_above( LOWER ) )
        self.assertTrue( AnnotationInspector( Optional[MIDDLE] ).is_indirectly_above( LOWER ) )
        self.assertTrue( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_above( LOWER ) )
        self.assertFalse( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_above( UPPER ) )
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_directly_above( LOWER ) )
        self.assertFalse( AnnotationInspector( Optional[MIDDLE] ).is_directly_above( LOWER ) )
        
        middle = MIDDLE()
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_viable_instance( middle ) )
        self.assertTrue( AnnotationInspector( Optional[UPPER] ).is_viable_instance( middle ) )
        self.assertTrue( AnnotationInspector( Union[LOWER, UPPER] ).is_viable_instance( middle ) )
        self.assertFalse( AnnotationInspector( Union[LOWER] ).is_viable_instance( middle ) )


class MathsHelper_test( unittest.TestCase ):
    def test_increment_mean( self ):
        from mhelper import maths_helper
        
        my_values = [1, 7, 23, 2, 6, 2, 6, 8, 42, 9, 3, 1, 0.2, 0.3]
        
        incremental_average = 0
        incremental_count = 0
        
        for index, new_value in enumerate( my_values ):
            incremental_average, incremental_count = maths_helper.increment_mean( incremental_average, incremental_count, new_value )
            true_average = sum( my_values[:index + 1] ) / (index + 1)
            self.assertAlmostEqual( incremental_average, true_average )
            self.assertEqual( incremental_count, index + 1 )


if __name__ == '__main__':
    unittest.main()



import traceback
import functools
from typing import Optional, Dict

from PyQt5.QtWidgets import QLineEdit

from mhelper import string_helper, ImplementationError, ansi
from mhelper.generics_helper import DECORATOR_FUNCTION, DECORATOR, DECORATED


def exceptToGui() -> DECORATOR_FUNCTION:
    """
    DECORATOR
    
    Same as `exqtSlot` but without the `pyqtSlot` bit.
    """
    
    
    def true_decorator( fn ) -> DECORATOR:
        @functools.wraps( fn )
        def fn_wrapper( *args, **kwargs ) -> DECORATED:
            try:
                return fn( *args, **kwargs )
            except Exception as ex:
                show_exception( args[0], "Error", ex )
        
        
        # noinspection PyTypeChecker
        return fn_wrapper
    
    
    return true_decorator


def exqtSlot( *decorator_args ) -> DECORATOR_FUNCTION:
    """
    DECORATOR
    
    `pyqtSlot` is problematic in that if an exception occurs the application will silently exit.
    This decorator replaces `pyqtSlot` by adding not only the decorator, but a simple error handler. 
    """
    
    
    def true_decorator( fn ) -> DECORATOR:
        from PyQt5.QtCore import pyqtSlot
        return pyqtSlot( *decorator_args )( exceptToGui()( fn ) )
    
    
    return true_decorator


def browse_dir_on_textbox( textbox ) -> bool:
    """
    Opens a file browser, using a textbox to retrieve and store the filename.
    :type textbox: QLineEdit
    """
    owner = textbox.window()
    
    from PyQt5.QtWidgets import QFileDialog
    dir_name = str( QFileDialog.getExistingDirectory( owner, "Select directory", textbox.text(), QFileDialog.DontUseNativeDialog ) )
    
    if not dir_name:
        return False
    
    textbox.setText( dir_name )
    
    return True


def browse_dir( owner, existing = None ):
    from PyQt5.QtWidgets import QFileDialog
    return str( QFileDialog.getExistingDirectory( owner, "Select directory", existing, QFileDialog.DontUseNativeDialog ) )


def browse_open_on_textbox( textbox:QLineEdit ) -> bool:
    """
    Opens a file browser, using a textbox to retrieve and store the filename.
    """
    from mhelper import file_helper
    
    owner = textbox.window()
    directory = file_helper.get_directory( textbox.text() )
    
    from PyQt5.QtWidgets import QFileDialog
    dir_name = str( QFileDialog.getOpenFileName( owner, "Select file", directory ) )
    
    if not dir_name:
        return False
    
    textbox.setText( dir_name )
    
    return True

def browse_save_on_textbox( textbox:QLineEdit, filter ) -> bool:
    """
    Opens a file browser, using a textbox to retrieve and store the filename.
    """
    from mhelper import file_helper
    
    owner = textbox.window()
    directory = file_helper.get_directory( textbox.text() )
    
    from PyQt5.QtWidgets import QFileDialog
    dir_name = str( QFileDialog.getSaveFileName( owner, "Select file", directory, filter, options = QFileDialog.DontUseNativeDialog ) )
    
    if not dir_name:
        return False
    
    textbox.setText( dir_name )
    
    return True


def browse_save( parent, filter ):
    from mhelper import file_helper
    
    from PyQt5.QtWidgets import QFileDialog
    file_name, file_filter = QFileDialog.getSaveFileName( parent, "Save", "", filter, options = QFileDialog.DontUseNativeDialog )
    
    if file_name:
        if not file_helper.get_extension( file_name ):
            file_filter = get_extension_from_filter( file_filter )
            
            file_name += file_filter
    
    return file_name or None


def browse_open( parent, filter ):
    from PyQt5.QtWidgets import QFileDialog
    file_name, file_filter = QFileDialog.getOpenFileName( parent, "Open", "", filter, options = QFileDialog.DontUseNativeDialog )
    
    return file_name or None


def get_extension_from_filter( file_filter ):
    """
    Gets the first extension from a filter of the form
    
    Data ( *.txt ) --> .txt
    Data (*.txt) --> .txt
    Data (*.txt, *.csv) --> .txt
    Data (*.txt *.csv) --> .txt
    etc.
    """
    file_filter = file_filter.split( "(", 1 )[1].strip( " *" )
    
    for x in " ,)":
        if x in file_filter:
            file_filter = file_filter.split( x, 1 )[0]
    return file_filter


# noinspection PyUnusedLocal
def show_exception( owner, message = None, exception = None, traceback_ = None ) -> None:
    if not traceback_:
        traceback_ = traceback.format_exc()
    
    if isinstance( message, BaseException ):
        exception = message
        message = "Error"
    
    from PyQt5.QtWidgets import QMessageBox
    msg = QMessageBox()
    msg.setIcon( QMessageBox.Critical )
    msg.setText( str( message ) )
    
    if isinstance( exception, BaseException ):
        print( "{}: {}".format( type( exception ).__name__, exception ) )
        print( traceback_ )
        msg.setInformativeText( str( exception ) )
        msg.setDetailedText( traceback_ )
        
        msg.exec_()


def to_check_state( value ):
    from PyQt5.QtCore import Qt
    if value is None:
        return Qt.PartiallyChecked
    elif value:
        return Qt.Checked
    else:
        return Qt.Unchecked


def from_check_state( value ):
    from PyQt5.QtCore import Qt
    if value == Qt.PartiallyChecked:
        return None
    elif value == Qt.Checked:
        return True
    elif value == Qt.Unchecked:
        return False
    else:
        from mhelper.exception_helper import SwitchError
        raise SwitchError( "from_check_state.value", value )


def move_treeview_items( source, destination, only_selected = True ):
    from PyQt5.QtWidgets import QTreeWidget
    assert isinstance( source, QTreeWidget )
    assert isinstance( destination, QTreeWidget )
    
    if only_selected:
        selected_items = source.selectedItems()
    else:
        selected_items = []
        
        for index in range( source.topLevelItemCount() ):
            selected_items.append( source.topLevelItem( index ) )
    
    for item in selected_items:
        index = source.indexOfTopLevelItem( item )
        source.takeTopLevelItem( index )
        destination.addTopLevelItem( item )


class AnsiHtmlLine:
    def __init__( self, code, back = None, fore = None, style = None, family = None ):
        """
        CONSTRUCTOR
        :param code:            ANSI code 
        :param back:            Back colour 
        :param fore:            Fore colour 
        :param style:           Font style 
        :param family:          Font family 
        """
        if isinstance( code, AnsiHtmlLine ):
            self.code = code.code
            self.back = code.back
            self.fore = code.fore
            self.style = code.style
            self.family = code.family
        else:
            self.code = code
            self.back = back
            self.fore = fore
            self.style = style
            self.family = family
    
    
    def to_style( self ):
        return 'background:{}; color:{}; font-style:{}; font-family:{}'.format( self.back, self.fore, self.style, self.family )
    
    
    def copy( self ):
        return AnsiHtmlLine( self.code, self.back, self.fore, self.style, self.family )


class AnsiHtmlScheme:
    def __init__( self, values ):
        values = [x if isinstance( x, AnsiHtmlLine ) else AnsiHtmlLine( *x ) for x in values]
        self.values: Dict[int, AnsiHtmlLine] = dict( (x.code, x) for x in values )
    
    
    def copy( self ) -> "AnsiHtmlScheme":
        return AnsiHtmlScheme( (x.copy() for x in self.values.values()) )
    
    
    def __getitem__( self, item: int ) -> AnsiHtmlLine:
        return self.values[item]
    
    
    def get_default( self ) -> AnsiHtmlLine:
        return self[ansi.CODE_INTERNAL_DEFAULT]


def ansi_scheme_dark( fg = "#FFFFFF", bg = "#000000", style = "normal", family = "sans-serif" ) -> AnsiHtmlScheme:
    """
    Creates a new ANSI scheme, defaulted to dark values.
    """
    return AnsiHtmlScheme( ((ansi.CODE_FORE_BLACK, "", "#000000", "", ""),
                            (ansi.CODE_FORE_RED, "", "#FF0000", "", ""),
                            (ansi.CODE_FORE_GREEN, "", "#00FF00", "", ""),
                            (ansi.CODE_FORE_YELLOW, "", "#FFFF00", "", ""),
                            (ansi.CODE_FORE_BLUE, "", "#0000FF", "", ""),
                            (ansi.CODE_FORE_MAGENTA, "", "#FF00FF", "", ""),
                            (ansi.CODE_FORE_CYAN, "", "#00FFFF", "", ""),
                            (ansi.CODE_FORE_WHITE, "", "#FFFFFF", "", ""),
                            (ansi.CODE_FORE_RESET, "", "*", "", ""),
                            (ansi.CODE_FORE_LIGHT_BLACK, "", "#808080", "", ""),
                            (ansi.CODE_FORE_LIGHT_RED, "", "#FF8080", "", ""),
                            (ansi.CODE_FORE_LIGHT_GREEN, "", "#80FF80", "", ""),
                            (ansi.CODE_FORE_LIGHT_YELLOW, "", "#FFFF80", "", ""),
                            (ansi.CODE_FORE_LIGHT_BLUE, "", "#8080FF", "", ""),
                            (ansi.CODE_FORE_LIGHT_MAGENTA, "", "#FF80FF", "", ""),
                            (ansi.CODE_FORE_LIGHT_CYAN, "", "#80FFFF", "", ""),
                            (ansi.CODE_FORE_LIGHT_WHITE, "", "#FFFFFF", "", ""),
                            (ansi.CODE_BACK_BLACK, "#000000", "", "", ""),
                            (ansi.CODE_BACK_RED, "#FF0000", "", "", ""),
                            (ansi.CODE_BACK_GREEN, "#00FF00", "", "", ""),
                            (ansi.CODE_BACK_YELLOW, "#FFFF00", "", "", ""),
                            (ansi.CODE_BACK_BLUE, "#0000FF", "", "", ""),
                            (ansi.CODE_BACK_MAGENTA, "#FF00FF", "", "", ""),
                            (ansi.CODE_BACK_CYAN, "#00FFFF", "", "", ""),
                            (ansi.CODE_BACK_WHITE, "#FFFFFF", "", "", ""),
                            (ansi.CODE_BACK_RESET, "*", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_BLACK, "#808080", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_RED, "#FF8080", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_GREEN, "#80FF80", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_YELLOW, "#FFFF80", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_BLUE, "#8080FF", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_MAGENTA, "#FF80FF", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_CYAN, "#80FFFF", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_WHITE, "#FFFFFF", "", "", ""),
                            (ansi.CODE_STYLE_RESET_ALL, "*", "*", "*", ""),
                            (ansi.CODE_STYLE_BRIGHT, "", "", "bold", ""),
                            (ansi.CODE_STYLE_DIM, "", "", "italic", ""),
                            (ansi.CODE_STYLE_NORMAL, "", "", "normal", ""),
                            (ansi.CODE_INTERNAL_DEFAULT, bg, fg, style, family),
                            (ansi.CODE_INTERNAL_QUOTE_START, "", "#00C0FF", "", "monospace"),
                            (ansi.CODE_INTERNAL_QUOTE_END, "", "*", "", "*"),
                            (ansi.CODE_INTERNAL_TABLE_START, "", "", "", "monospace"),
                            (ansi.CODE_INTERNAL_TABLE_END, "", "", "", "*"),
                            ) )


def ansi_scheme_light( fg = "#000000", bg = "#FFFFFF" ) -> AnsiHtmlScheme:
    """
    Creates a new ANSI scheme, defaulted to light values.
    """
    return AnsiHtmlScheme( ((ansi.CODE_FORE_BLACK, "", "#000000", "", ""),
                            (ansi.CODE_FORE_RED, "", "#C00000", "", ""),
                            (ansi.CODE_FORE_GREEN, "", "#00C000", "", ""),
                            (ansi.CODE_FORE_YELLOW, "", "#C0C000", "", ""),
                            (ansi.CODE_FORE_BLUE, "", "#0000C0", "", ""),
                            (ansi.CODE_FORE_MAGENTA, "", "#C000C0", "", ""),
                            (ansi.CODE_FORE_CYAN, "", "#00C0C0", "", ""),
                            (ansi.CODE_FORE_WHITE, "", "#C0C0C0", "", ""),
                            (ansi.CODE_FORE_RESET, "", "*", "", ""),
                            (ansi.CODE_FORE_LIGHT_BLACK, "", "#C0C0C0", "", ""),
                            (ansi.CODE_FORE_LIGHT_RED, "", "#FF0000", "", ""),
                            (ansi.CODE_FORE_LIGHT_GREEN, "", "#00FF00", "", ""),
                            (ansi.CODE_FORE_LIGHT_YELLOW, "", "#FFFF00", "", ""),
                            (ansi.CODE_FORE_LIGHT_BLUE, "", "#0000FF", "", ""),
                            (ansi.CODE_FORE_LIGHT_MAGENTA, "", "#FF00FF", "", ""),
                            (ansi.CODE_FORE_LIGHT_CYAN, "", "#00FFFF", "", ""),
                            (ansi.CODE_FORE_LIGHT_WHITE, "", "#FFFFFF", "", ""),
                            (ansi.CODE_BACK_BLACK, "#000000", "", "", ""),
                            (ansi.CODE_BACK_RED, "#C00000", "", "", ""),
                            (ansi.CODE_BACK_GREEN, "#00C000", "", "", ""),
                            (ansi.CODE_BACK_YELLOW, "#C0C000", "", "", ""),
                            (ansi.CODE_BACK_BLUE, "#0000C0", "", "", ""),
                            (ansi.CODE_BACK_MAGENTA, "#C000C0", "", "", ""),
                            (ansi.CODE_BACK_CYAN, "#00C0C0", "", "", ""),
                            (ansi.CODE_BACK_WHITE, "#C0C0C0", "", "", ""),
                            (ansi.CODE_BACK_RESET, "*", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_BLACK, "#C0C0C0", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_RED, "#FF0000", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_GREEN, "#00FF00", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_YELLOW, "#FFFF00", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_BLUE, "#0000FF", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_MAGENTA, "#FF00FF", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_CYAN, "#00FFFF", "", "", ""),
                            (ansi.CODE_BACK_LIGHT_WHITE, "#FFFFFF", "", "", ""),
                            (ansi.CODE_STYLE_RESET_ALL, "*", "*", "*", ""),
                            (ansi.CODE_STYLE_BRIGHT, "", "", "bold", ""),
                            (ansi.CODE_STYLE_DIM, "", "", "italic", ""),
                            (ansi.CODE_STYLE_NORMAL, "", "", "normal", ""),
                            (ansi.CODE_INTERNAL_DEFAULT, bg, fg, "normal", "sans-serif"),
                            (ansi.CODE_INTERNAL_QUOTE_START, "", "#0080C0", "", "Consolas, monospace"),
                            (ansi.CODE_INTERNAL_QUOTE_END, "", "*", "", "*"),
                            (ansi.CODE_INTERNAL_TABLE_START, "", "", "", "Consolas, monospace"),
                            (ansi.CODE_INTERNAL_TABLE_END, "", "", "", "*")
                            ) )


def ansi_to_html( text: Optional[str], lookup: Optional[AnsiHtmlScheme] = None, *, debug: bool = False ) -> str:
    """
    Converts ANSI text to HTML.
    
    :param text:            Text to convert 
    :param lookup:          An `AnsiHtmlScheme` object defining the colours to use, or `None` for the default (dark).
                            See functions `ansi_scheme_dark` and `ansi_scheme_light`. 
    :param debug:           When `True`, prints the results to a temporary file. 
    :return:                The HTML. 
    """
    if text is None:
        return ""
    
    from mhelper.string_parser import StringParser
    html = []
    
    lookup = lookup.values if lookup is not None else ansi_scheme_dark().values
    default_style = lookup[ansi.CODE_INTERNAL_DEFAULT]
    current_style = AnsiHtmlLine( default_style )
    
    text = string_helper.highlight_quotes( text, "`", "`", "\033[-2m", "\033[-3m" )
    text = string_helper.highlight_quotes( text, "┌", "┘", "\033[-4m", "\033[-5m" )
    text = string_helper.highlight_quotes( text, "«", "»", "\033[-2m", "\033[-3m" )
    
    text = text.replace( "\n* ", "\n• " )
    text = text.replace( "\n", '<br/>' )
    text = text.replace( " ", ' ' )
    
    parser = StringParser( text )
    
    html.append( '<span style="background:{}; color:{}; font-style:{}; font-family:{};">'.format( current_style.back, current_style.fore, current_style.style, current_style.family ) )
    iterations = 0
    
    ANSI_ESCAPE = '\033['
    
    while not parser.end():
        iterations += 1
        
        if iterations >= 10000:
            raise ImplementationError( "Possible infinite loop in mhelper.qt_gui_helper.ansi_to_html. Internal error. Possible solutions: Use a shorter string. Check the lookup syntax." )
        
        html.append( parser.read_past( ANSI_ESCAPE ) )
        
        if parser.end():
            break
        
        code_str = parser.read_past( "m" )
        
        try:
            code = int( code_str )
        except:
            code = None
        
        if code is not None:
            new_style = lookup[code]
            
            for attr in "fore", "back", "family", "style":
                new = getattr( new_style, attr )
                if new:
                    if new == "*":
                        setattr( current_style, attr, getattr( default_style, attr ) )
                    else:
                        setattr( current_style, attr, new )
            
            html.append( '<span style="{}">'.format( current_style.to_style() ) )
    
    html.append( '</span>' )
    
    result = "".join( html )
    
    if debug:
        txx = ["ANSI", text, "HTML", result]
        from mhelper import file_helper
        file_helper.write_all_text( "temp-del-me.txt", "\n".join( txx ) )
    
    return result


__author__ = "Martin Rusilowicz"


class QtMutex:
    """
    A `QMutex` wrapped in Python.
    
    This is a `QMutex` and requires QT.
    - It is only for the GUI!
    - - Plugins should *not* use this!
    
    Usage:

        ```
        m = QtMutex()
        
        with m:
           . . .
           
        ```
           
    """
    
    
    def __init__( self ):
        from PyQt5.QtCore import QMutex
        self._mutex = QMutex()
    
    
    def __enter__( self ):
        self._mutex.lock()
        return self
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        self._mutex.unlock()


def suppress_debugging():
    # Disable PyQT logging to console since we use console for other stuff, plus, it's irritating
    # noinspection PyUnusedLocal
    def __message_handler( msg_type, msg_log_context, msg_string ):
        pass
    
    
    from PyQt5 import QtCore
    QtCore.qInstallMessageHandler( __message_handler )



class UtfSub:
    def __init__( self, *replacement ):
        if len( replacement ) == 1:
            replacement = replacement[0]
        
        if isinstance( replacement, tuple ):
            replacement = "-----------------" + replacement[0] + "-------" + replacement[1] + "------" + replacement[2] + "-----"
        
        if len( replacement ) != len( _normal_str ):
            raise ValueError( "Replacement text {} but must be {} characters long:\n{}\n{}".format( len( replacement ), len( normal ), replacement, normal ) )
        
        self.replacement = replacement
    
    
    def __call__( self, text ):
        replacement = self.replacement
        r = []
        
        for c in text:
            index = ord( c ) - 31
            
            if index > len( replacement ):
                raise ValueError( "Index ({}, {}) above replacement length ({}).".format( c, index, len( replacement ) ) )
            
            if replacement[index]=="-":
                r.append( c )
            else:
                r.append( replacement[index] )
        
        return "".join( r )


_normal_str = " !\"# $%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
normal = UtfSub( _normal_str )

serif_bold = UtfSub( "𝟎𝟏𝟐𝟑𝟒𝟓𝟔𝟕𝟖𝟗", "𝐀𝐁𝐂𝐃𝐄𝐅𝐆𝐇𝐈𝐉𝐊𝐋𝐌𝐍𝐎𝐏𝐐𝐑𝐒𝐓𝐔𝐕𝐖𝐗𝐘𝐙", "𝐚𝐛𝐜𝐝𝐞𝐟𝐠𝐡𝐢𝐣𝐤𝐥𝐦𝐧𝐨𝐩𝐪𝐫𝐬𝐭𝐮𝐯𝐰𝐱𝐲𝐳" )
serif_bold_italic = UtfSub( "0123456789", "𝑨𝑩𝑪𝑫𝑬𝑭𝑮𝑯𝑰𝑱𝑲𝑳𝑴𝑵𝑶𝑷𝑸𝑹𝑺𝑻𝑼𝑽𝑾𝑿𝒀𝒁", "𝒂𝒃𝒄𝒅𝒆𝒇𝒈𝒉𝒊𝒋𝒌𝒍𝒎𝒏𝒐𝒑𝒒𝒓𝒔𝒕𝒖𝒗𝒘𝒙𝒚𝒛" )

fraktur = UtfSub( "0123456789", "𝕬𝕭𝕮𝕯𝕰𝕱𝕲𝕳𝕴𝕵𝕶𝕷𝕸𝕹𝕺𝕻𝕼𝕽𝕾𝕿𝖀𝖁𝖂𝖃𝖄𝖅", "𝖆𝖇𝖈𝖉𝖊𝖋𝖌𝖍𝖎𝖏𝖐𝖑𝖒𝖓𝖔𝖕𝖖𝖗𝖘𝖙𝖚𝖛𝖜𝖝𝖞𝖟" )
script = UtfSub( "0123456789", "𝓐𝓑𝓒𝓓𝓔𝓕𝓖𝓗𝓘𝓙𝓚𝓛𝓜𝓝𝓞𝓟𝓠𝓡𝓢𝓣𝓤𝓥𝓦𝓧𝓨𝓩", "𝓪𝓫𝓬𝓭𝓮𝓯𝓰𝓱𝓲𝓳𝓴𝓵𝓶𝓷𝓸𝓹𝓺𝓻𝓼𝓽𝓾𝓿𝔀𝔁𝔂𝔃" )
doublestruck = UtfSub( "𝟘𝟙𝟚𝟛𝟜𝟝𝟞𝟟𝟠𝟡", "𝔸𝔹ℂ𝔻𝔼𝔽𝔾ℍ𝕀𝕁𝕂𝕃𝕄ℕ𝕆ℙℚℝ𝕊𝕋𝕌𝕍𝕎𝕏𝕐ℤ", "𝕒𝕓𝕔𝕕𝕖𝕗𝕘𝕙𝕚𝕛𝕜𝕝𝕞𝕟𝕠𝕡𝕢𝕣𝕤𝕥𝕦𝕧𝕨𝕩𝕪𝕫" )
monospaced = UtfSub( "𝟶𝟷𝟸𝟹𝟺𝟻𝟼𝟽𝟾𝟿", "𝙰𝙱𝙲𝙳𝙴𝙵𝙶𝙷𝙸𝙹𝙺𝙻𝙼𝙽𝙾𝙿𝚀𝚁𝚂𝚃𝚄𝚅𝚆𝚇𝚈𝚉", "𝚊𝚋𝚌𝚍𝚎𝚏𝚐𝚑𝚒𝚓𝚔𝚕𝚖𝚗𝚘𝚙𝚚𝚛𝚜𝚝𝚞𝚟𝚠𝚡𝚢𝚣" )

# Math sans-serif
sans = UtfSub( "𝟢𝟣𝟤𝟥𝟦𝟧𝟨𝟩𝟪𝟫", "𝖠𝖡𝖢𝖣𝖤𝖥𝖦𝖧𝖨𝖩𝖪𝖫𝖬𝖭𝖮𝖯𝖰𝖱𝖲𝖳𝖴𝖵𝖶𝖷𝖸𝖹", "𝖺𝖻𝖼𝖽𝖾𝖿𝗀𝗁𝗂𝗃𝗄𝗅𝗆𝗇𝗈𝗉𝗊𝗋𝗌𝗍𝗎𝗏𝗐𝗑𝗒𝗓" )
sans_bold = UtfSub( "𝟬𝟭𝟮𝟯𝟰𝟱𝟲𝟳𝟴𝟵", "𝗔𝗕𝗖𝗗𝗘𝗙𝗚𝗛𝗜𝗝𝗞𝗟𝗠𝗡𝗢𝗣𝗤𝗥𝗦𝗧𝗨𝗩𝗪𝗫𝗬𝗭", "𝗮𝗯𝗰𝗱𝗲𝗳𝗴𝗵𝗶𝗷𝗸𝗹𝗺𝗻𝗼𝗽𝗾𝗿𝘀𝘁𝘂𝘃𝘄𝘅𝘆𝘇" )
sans_bold_italic = UtfSub( "0123456789", "𝘼𝘽𝘾𝘿𝙀𝙁𝙂𝙃𝙄𝙅𝙆𝙇𝙈𝙉𝙊𝙋𝙌𝙍𝙎𝙏𝙐𝙑𝙒𝙓𝙔𝙕", "𝙖𝙗𝙘𝙙𝙚𝙛𝙜𝙝𝙞𝙟𝙠𝙡𝙢𝙣𝙤𝙥𝙦𝙧𝙨𝙩𝙪𝙫𝙬𝙭𝙮𝙯" )
sans_italic = UtfSub( "0123456789", "𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡", "𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻" )

# Parentheses
parenthesis = UtfSub( "0⑴⑵⑶⑷⑸⑹⑺⑻⑼", "⒜⒝⒞⒟⒠⒡⒢⒣⒤⒥⒦⒧⒨⒩⒪⒫⒬⒭⒮⒯⒰⒱⒲⒳⒴⒵", "⒜⒝⒞⒟⒠⒡⒢⒣⒤⒥⒦⒧⒨⒩⒪⒫⒬⒭⒮⒯⒰⒱⒲⒳⒴⒵" )

# Circled 
circled = UtfSub( "0①②③④⑤⑥⑦⑧⑨", "ⒶⒷⒸⒹⒺⒻⒼⒽⒾⒿⓀⓁⓂⓃⓄⓅⓆⓇⓈⓉⓊⓋⓌⓍⓎⓏ", "ⓐⓑⓒⓓⓔⓕⓖⓗⓘⓙⓚⓛⓜⓝⓞⓟⓠⓡⓢⓣⓤⓥⓦⓧⓨⓩ" )
circled_black = UtfSub( "⓿123456789", "🅐🅑🅒🅓🅔🅕🅖🅗🅘🅙🅚🅛🅜🅝🅞🅟🅠🅡🅢🅣🅤🅥🅦🅧🅨🅩", "🅐🅑🅒🅓🅔🅕🅖🅗🅘🅙🅚🅛🅜🅝🅞🅟🅠🅡🅢🅣🅤🅥🅦🅧🅨🅩" )

# Square
square = UtfSub( "0123456789", "🄰🄱🄲🄳🄴🄵🄶🄷🄸🄹🄺🄻🄼🄽🄾🄿🅀🅁🅂🅃🅄🅅🅆🅇🅈🅉", "🄰🄱🄲🄳🄴🄵🄶🄷🄸🄹🄺🄻🄼🄽🄾🄿🅀🅁🅂🅃🅄🅅🅆🅇🅈🅉" )
square_black = UtfSub( "0123456789", "🅰🅱🅲🅳🅴🅵🅶🅷🅸🅹🅺🅻🅼🅽🅾🅿🆀🆁🆂🆃🆄🆅🆆🆇🆈🆉", "🅰🅱🅲🅳🅴🅵🅶🅷🅸🅹🅺🅻🅼🅽🅾🅿🆀🆁🆂🆃🆄🆅🆆🆇🆈🆉" )

# Regional indicator
regional = UtfSub( "0123456789", "🇦🇧🇨🇩🇪🇫🇬🇭🇮🇯🇰🇱🇲🇳🇴🇵🇶🇷🇸🇹🇺🇻🇼🇽🇾🇿", "🇦🇧🇨🇩🇪🇫🇬🇭🇮🇯🇰🇱🇲🇳🇴🇵🇶🇷🇸🇹🇺🇻🇼🇽🇾🇿" )

# Full-width
fullwidth = UtfSub( "０１２３４５６７８９", "ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ", "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ" )

# Small caps
smallcaps = UtfSub( "0123456789", "ABCDEFGHIJKLMNOPQRSTUVWXYZ", "ᴀʙᴄᴅᴇꜰɢʜɪᴊᴋʟᴍɴᴏᴩqʀꜱᴛᴜᴠᴡxʏᴢ" )

# Sub/super
superscript = UtfSub( "⁰¹²³⁴⁵⁶⁷⁸⁹", "ᴬᴮ-ᴰᴱ-ᴳᴴᴵᴶᴷᴸᴹᴺᴼᴾ---ᵀᵁⱽᵂ---", "ᵃᵇᶜᵈᵉᶠᵍʰⁱʲᵏˡᵐⁿᵒᵖ-ʳˢᵗᵘᵛʷˣʸᶻ" )
subscript = UtfSub( "₀₁₂₃₄₅₆₇₈₉", "--------------------------", "ₐ---ₑ---ᵢⱼ----ₒ--ᵣ--ᵤᵥ-ₓ--" )

# Misc
greek = UtfSub( "0123456789", "ΑΒΧΔΕΦΓΗΙJΚΛΜΝΟΠΘΡΣΤΨVΩΞΥΖ", "αβχδεφγηιjκλμνοπθρστψvωξυζ" )



from distutils.core import setup


setup( name = "mhelper",
       url = "https://bitbucket.org/mjr129/mhelper",
       version = "1.0.1.41",
       description = "Includes a collection of utility functions.",
       author = "Martin Rusilowicz",
       license = "https://www.gnu.org/licenses/agpl-3.0.html",
       packages = ["mhelper",
                   "mhelper_qt",
                   "mdbg"],
       python_requires = ">=3.6",
       install_requires = ["PyQt5",
                           "py-flags",
                           "jsonpickle"]
       )



"""
Functions for formatting stuff using ANSI codes and/or esoteric UNICODE characters.
"""

from typing import Union, Iterable, cast

from colorama import Back, Fore, Style

from mhelper import ansi, ansi_helper, exception_helper, string_helper


def format_source( text: str,
                   keywords: Iterable[str],
                   variables: Iterable[str] ) -> str:
    """
    Prints source text, highlighting keywords and variables, and prefixing line numbers
    
    :param text:        Text to print
    :param keywords:    Keywords to highlight
    :param variables:   Variables to highlight
    :return:            Nothing
    """
    r = []
    
    text = text.split( "\n" )
    
    for i, line in enumerate( text ):
        prefix = Back.LIGHTBLACK_EX + Fore.BLACK + " " + str( i ).rjust( 4 ) + " " + Style.RESET_ALL + " "
        
        line = string_helper.highlight_words( line, keywords, Style.RESET_ALL + Fore.GREEN, Style.RESET_ALL )
        line = string_helper.highlight_words( line, variables, Style.RESET_ALL + Fore.CYAN, Style.RESET_ALL )
        
        r.append( prefix + line )
    
    return "\n".join( r )


def format_traceback( exception: Union[BaseException, str],
                      traceback_ = None,
                      warning = False,
                      wordwrap = 0 ) -> str:
    """
    Formats a traceback.
    
    :param exception:       Exception to display 
    :param traceback_:      Traceback text (leave as `None` to get the system traceback) 
    :param warning:         Set to `True` to use warning, rather than error, colours 
    :param wordwrap:        Set to the wordwrap width. 
    :return:                ANSI containing string  
    """
    output_list = []
    
    wordwrap = wordwrap or 140
    INTERIOR_WIDTH = wordwrap - 4
    
    if warning:
        ⅎMAIN = Style.RESET_ALL + Back.YELLOW + Fore.BLACK
        ⅎHIGH = Style.RESET_ALL + Style.BRIGHT + Back.YELLOW + Fore.BLACK
        ⅎBORDER = Back.LIGHTYELLOW_EX + Fore.BLACK
        ⅎNORMAL = Fore.CYAN
        ⅎCODE = Fore.CYAN + Style.DIM
        ⅎLINE = Style.DIM + Fore.CYAN
        ⅎFILE = Fore.LIGHTCYAN_EX
    else:
        ⅎMAIN = Style.RESET_ALL + Back.WHITE + Fore.RED
        ⅎHIGH = Style.RESET_ALL + Back.WHITE + Fore.BLACK + ansi.ITALIC
        ⅎBORDER = Back.RED + Fore.WHITE
        ⅎNORMAL = Style.RESET_ALL + Back.WHITE + Fore.RED
        ⅎCODE = Style.RESET_ALL + Back.WHITE + Fore.BLACK
        ⅎLINE = Style.RESET_ALL + Back.WHITE + Fore.BLUE + ansi.DIM
        ⅎFILE = Style.RESET_ALL + Back.WHITE + Fore.BLUE
    
    S_V, S_TL, S_H, S_TR, S_VL, S_VR, S_BL, S_BR = "│┌─┐├┤└┘"
    
    LBORD = ⅎBORDER + S_V + ⅎMAIN + " "
    RBORD = ⅎMAIN + " " + ⅎBORDER + S_V + Style.RESET_ALL
    
    output_list.append( ⅎBORDER + S_TL + S_H * (wordwrap - 2) + S_TR + Style.RESET_ALL )
    
    if isinstance( exception, BaseException ):
        ex = exception
        
        while ex:
            if ex is not exception:
                output_list.append( LBORD + " ---CAUSED BY---".ljust( INTERIOR_WIDTH ) + RBORD )
            
            output_list.append( LBORD + ("⦂" + type( ex ).__name__ + "⦂").ljust( INTERIOR_WIDTH ) + RBORD )
            ex_text = ⅎMAIN + string_helper.highlight_quotes( str( ex ), "«", "»", ⅎHIGH, ⅎMAIN )
            
            for line in ansi_helper.wrap( ex_text, INTERIOR_WIDTH ):
                line = ansi_helper.ljust( line, INTERIOR_WIDTH )
                output_list.append( LBORD + line + RBORD )
                if len( output_list ) > 20:
                    output_list.append( LBORD + ansi_helper.ljust( "...(trunacated)...", INTERIOR_WIDTH ) + RBORD )
                    break
            
            ex = ex.__cause__
    
    else:
        output_list.append( LBORD + str( exception ).ljust( INTERIOR_WIDTH ) + RBORD )
    
    if not traceback_:
        traceback_ = exception_helper.full_traceback()
    
    lines = traceback_.split( "\n" )
    
    for i, line in enumerate( lines ):
        print_line = line.strip()
        
        if print_line.startswith( "File " ):
            print_line = ⅎLINE + string_helper.highlight_regex( print_line, "\\/([^\\/]*)\"", ⅎFILE, ⅎLINE )
            print_line = ⅎLINE + string_helper.highlight_regex( print_line, "line ([0-9]*),", ⅎFILE, ⅎLINE )
            print_line = ⅎLINE + string_helper.highlight_regex( print_line, "in (.*)$", ⅎCODE, ⅎLINE )
            print_line = ansi_helper.fix_width( print_line, INTERIOR_WIDTH )
            output_list.append( LBORD + print_line + RBORD )
        elif line.startswith( " " ):
            print_line = ansi_helper.fix_width( print_line, INTERIOR_WIDTH )
            output_list.append( LBORD + ansi_helper.fix_width( "    " + ⅎCODE + print_line, INTERIOR_WIDTH ) + RBORD )
        elif line.startswith( "*" ):
            c = wordwrap - len( print_line ) - 6
            output_list.append( ⅎBORDER + S_VL + cast( str, S_H * 4 ) + print_line + S_H * c + S_VR + Style.RESET_ALL )
        else:
            print_line = ansi_helper.fix_width( print_line, INTERIOR_WIDTH )
            output_list.append( LBORD + ⅎNORMAL + print_line + RBORD )
    
    output_list.append( ⅎBORDER + S_BL + S_H * (wordwrap - 2) + S_BR + Style.RESET_ALL )
    
    return "\n".join( output_list )


def format_two_columns( *,
                        left_margin: int,
                        centre_margin: int,
                        right_margin: int,
                        left_text: str,
                        right_text: str,
                        left_prefix: str = "",
                        right_prefix: str = "",
                        left_suffix: str = "",
                        right_suffix: str = "", ):
    """
    Formats a box.
    :param *: 
    :param left_margin:     Width of left margin 
    :param centre_margin:   Width of centre margin 
    :param right_margin:    Width of right margin 
    :param left_text:       Text in left column 
    :param right_text:      Text in right column 
    :param left_prefix:     Text added to left lines
    :param right_prefix:    Text added to right lines
    :param left_suffix:     Text added to left lines
    :param right_suffix:    Text added to right lines
    :return: 
    """
    r = []
    left_space = centre_margin - left_margin - 1
    right_space = right_margin - centre_margin
    
    left_lines = list( left_prefix + x + left_suffix for x in ansi_helper.wrap( left_text, left_space, pad = True ) )
    right_lines = list( right_prefix + x + right_suffix for x in ansi_helper.wrap( right_text, right_space ) )
    num_lines = max( len( left_lines ), len( right_lines ) )
    
    for i in range( num_lines ):
        left = left_lines[i] if i < len( left_lines ) else " " * left_space
        right = right_lines[i] if i < len( right_lines ) else " " * right_space
        
        text = (" " * left_margin) + left + Style.RESET_ALL + " " + right + Style.RESET_ALL
        r.append( text )
    
    return "\n".join( r )



"""
Defines the `@itemproperty` decorator and its supporting classes.
"""


def itemproperty( f ):
    """
    Like the `@property` decorator, but creates an indexable property.
    
    Note that the _setter_ is called `itemsetter` and not `setter`.
    (This is to avoid PyLint or IDEs being confused as they make certain assumptions about `.setter` decorators.)
    
    ```
    class MyClass:
        @itemproperty
        def eggs( item ):
            return self._eggs[item]
            
        @eggs.itemsetter
        def eggs( item, value ):
           self._eggs[item] = value 
    
    m = MyClass()
    m.eggs["spam"] = "beans"
    print( m.eggs["spam"] )
    ``` 
    """
    return _ItemPropertyClass( f )


class _ItemPropertyClass:
    def __init__( self, getter ):
        self._getter = getter
        self._setter = None
    
    
    def __get__( self, instance, owner ):
        return _BoundItemPropertyClass( instance, self )
    
    
    def itemsetter( self, setter ):
        self._setter = setter
        return self


class _BoundItemPropertyClass:
    def __init__( self, self2, class_: _ItemPropertyClass ):
        self.__self2 = self2
        self.__class = class_
    
    
    def __getitem__( self, item ):
        return self.__class._getter( self.__self2, item )
    
    
    def __setitem__( self, key, value ):
        return self.__class._setter( self.__self2, key, value )



import subprocess
import traceback
from typing import Iterable, Union, Optional

import itertools


TType = Union[type, Iterable[type]]
"""A type, or a collection of types"""


class NotSupportedError( Exception ):
    """
    Since `NotImplementedError` looks like an abstract-base-class error to the IDE, `NotSupportedError` provides a more explicit alternative.
    """
    pass


class LogicError( Exception ):
    """
    Signifies a logical error in the subroutine which generally isn't the caller's fault.
    """
    pass


ImplementationError = LogicError
"""Alias for LogicError"""


class NotFoundError( Exception ):
    """
    Like FileNotFound error, but when applied to something other than files.
    """
    pass


class SwitchError( Exception ):
    """
    An error selecting the case of a switch.
    """
    
    
    def __init__( self, name: str, value: object, *, instance: bool = False, details: Optional[str] = None ):
        """
        CONSTRUCTOR
        
        :param name:        Name of the switch 
        :param value:       Value passed to the switch 
        :param instance:    Set to indicate the switch is on the type of value (`type(value)`)
        :param details:     Additional message to append to the error text. 
        """
        if details is not None:
            details = " Further details: {}".format( details )
        
        if instance:
            super().__init__( "The switch on the type of «{}» does not recognise the value «{}» of type «{}».{}".format( name, value, type( value ), details ) )
        else:
            super().__init__( "The switch on «{}» does not recognise the value «{}» of type «{}».{}".format( name, value, type( value ), details ) )


class SubprocessError( Exception ):
    """
    Raised when the result of calling a subprocess indicates an error.
    """
    pass


def add_details( exception: Exception, **kwargs ) -> None:
    """
    Attaches arbitrary information to an exception.
    
    :param exception:   Exception 
    :param kwargs:      Information to attach
    """
    args = list( exception.args )
    
    message = create_details_message( **kwargs )
    
    if len( args ) > 0 and isinstance( args[0], str ):
        args[0] += message
    else:
        args.append( message )
    
    exception.args = tuple( args )


def create_details_message( **kwargs ):
    from mhelper import string_helper
    
    result = [""]
    
    lk = 1
    lt = 1
    
    for k, v in kwargs.items():
        lk = max( len( str( k ) ), lk )
        lt = max( len( string_helper.type_name( v ) ), lt )
    
    for k, v in kwargs.items():
        result.append( "--> {0} ({1}) = «{2}»".format( str( k ).ljust( lk ), string_helper.type_name( v ).ljust( lt ), v ) )
    
    return "\n".join( result )


def assert_type( name, value, type ):
    if not isinstance( value, type ):
        from mhelper.string_helper import type_name
        raise TypeError( "`{0}` should be of type `{1}`, but it is a `{2}` with value `{3}`.".format( name, type.__name__, type_name( value ), value ) )


def exception_to_string( ex: BaseException ):
    result = []
    
    while ex:
        result.append( str( ex ) )
        ex = ex.__cause__
    
    return "\n---CAUSED BY---\n".join( result )


def run_subprocess( command: str ) -> None:
    """
    Runs a subprocess, raising `SubprocessError` if the error code is set.
    """
    status = subprocess.call( command, shell = True )
    
    if status:
        raise SubprocessError( "The command «{}» exited with error code «{}». If available, checking the console output may provide more details.".format( command, status ) )


def format_types( type_: TType ) -> str:
    if isinstance( type_, type ):
        return str( type_ )
    else:
        from mhelper import string_helper
        return string_helper.join_ex( type_, delimiter = ", ", last_delimiter = " or ", formatter = "«{}»" )


def assert_instance( name: str, value: object, type_: TType ):
    if isinstance( type_, type ):
        type_ = (type_,)
    
    if not any( isinstance( value, x ) for x in type_ ):
        raise TypeError( instance_message( name, value, type_ ) )


def assert_instance_or_none( name: str, value: object, type_: type ):
    if isinstance( type_, type ):
        type_ = (type_,)
    
    type_ = list( itertools.chain( type_, (type( None ),) ) )
    
    assert_type( name, value, type_ )


def instance_message( name: str, value: object, type_: TType ) -> str:
    """
    Creates a suitable message describing a type error.
    :param name:        Name 
    :param value:       Value 
    :param type_:       Expected type 
    :return:            The message
    """
    return "The value of «{}», which is «{}», should be of type {}, but it's not, it's a «{}».".format( name, value, format_types( type_ ), type( value ) )


def full_traceback():
    return "**** Handler Traceback ****\n" + current_stack_text() + "\n**** Error traceback ****\n" + traceback.format_exc()


def current_stack_text():
    return "\n".join( x.strip() for x in traceback.format_stack() )


def type_error( name: str, value: object, type_: TType ) -> None:
    """
    Raises a `TypeError` with an appropriate message.
    
    :param name:        Name 
    :param value:       Value 
    :param type_:       Expected type 
    :except:            TypeError
    """
    raise TypeError( instance_message( name, value, type_ ) )



from mhelper import ansi


class QfState:
    def __init__( self, formatter ):
        assert isinstance( formatter, QuickFormatter )
        
        self.formatter = formatter
        
        self.stack = []
        self.r = []
    
    
    def format( self, text: str ):
        l = len( text )
        i = 0
        li = 0
        
        while i < l:
            c = text[i]
            
            if c == "<":
                if li != i:
                    self.append( text[li:i] )
                
                j = i + 1
                while text[j] != ">" and j < l:
                    j += 1
                
                self.append( self.process( text[i + 1:j] ) )
                i = j + 1
                li = i
            else:
                i += 1
        
        if li != i:
            self.append( text[li:i] )
        
        return "".join( r )
    
    
    def process( self, p ):
        if p == "br" or p == "br/":
            return "\n"
        elif p.startswith( "/" ):
            self.pop()
        elif p == "h1":
            self.append()
            self.push( ansi.FORE_BRIGHT_WHITE + ansi.BACK_BLUE,  )
        elif p == "h2":
            self.push( ansi.FORE_BRIGHT_WHITE + ansi.BACK_BLUE )
            
        return "*" + p + "*"
    
    
    def append( self, param ):
        self.r.append( param )
    
    
    def push( self, fg ):
        self.stack.append( fg )
        self.append( fg )
    
    
    def pop( self ):
        self.append( self.stack.pop() )


class QuickFormatter:
    def __init__( self ):
        pass
    
    
    def format( self, text: str ):
        return QfState( self ).format( text )


STANDARD = QuickFormatter()

print( "[" + STANDARD.format( "hello <world> oh world" ) + "]" )
